import atexit
import logging
import os
import socket
import struct
import threading
import time
from abc import ABCMeta, abstractmethod
from itertools import cycle
from pathlib import Path
from typing import Final, Mapping, Tuple, Union
from weakref import WeakSet

import flufl.lock as fl

from quel_ic_config.box_lock import BoxLockError
from quel_ic_config.exstickge_proxy import LsiKindId, _ExstickgeProxyBase
from quel_ic_config.quel_config_common import _DEFAULT_LOCK_DIRECTORY

logger = logging.getLogger(__name__)


class AbstractLockKeeper(threading.Thread, metaclass=ABCMeta):
    _DEFAULT_LOOP_WAIT: Final[float] = 0.25  # [s]
    _DEFAULT_RELEASE_TIMEOUT: Final[float] = 2.5
    _DEFAULT_LOCK_EXPIRATION: float

    _clients: WeakSet["AbstractLockKeeper"] = WeakSet()
    _clients_lock: threading.RLock = threading.RLock()

    @classmethod
    def release_lock_all(cls):
        with cls._clients_lock:
            # Notes: just for speeding up
            for c in cls._clients:
                c._to_release = True
            for c in cls._clients:
                c.deactivate()

    def __init__(self, *, target: tuple[str, int], loop_wait: float = _DEFAULT_LOOP_WAIT):
        super().__init__()
        self._target: tuple[str, int] = target
        self._loop_wait = loop_wait
        self._lock_timestamp: float = 0.0
        self._to_release: bool = False
        # Note: must be daemon to allow invocation of release_lock_all at exit
        #       even if some lock_keepers are still alive.
        self.daemon = True

    def _register_self(self):
        with self._clients_lock:
            if not self.has_lock:
                raise RuntimeError(f"try to register proxy object for {self._target[0]} which doesn't have lock")

            for c in self._clients:
                if c._target[0] == self._target[0] and c.has_lock:
                    # Notes: this doesn't happen usually, just in case.
                    raise RuntimeError(f"duplicated proxy object for {self._target[0]}")
            else:
                self._clients.add(self)

    def _unregister_self(self):
        with self._clients_lock:
            if not self.has_lock:
                try:
                    self._clients.remove(self)
                except KeyError:
                    pass
            else:
                raise RuntimeError(f"try to unregister live lock for {self._target[0]}")

    @property
    @abstractmethod
    def has_lock(self) -> bool: ...

    @abstractmethod
    def _take_lock(self) -> bool: ...

    @abstractmethod
    def _keep_lock(self) -> bool: ...

    @abstractmethod
    def _release_lock(self) -> None: ...

    def activate(self) -> bool:
        if self._take_lock():
            self._lock_timestamp = time.perf_counter()
            self._register_self()
            self.start()
            return True
        else:
            return False

    def run(self):
        while not self._to_release:
            if self.has_lock:
                cur = time.perf_counter()
                if (cur - self._lock_timestamp) > self._DEFAULT_LOCK_EXPIRATION * 0.5:
                    if self._keep_lock():
                        self._lock_timestamp = cur
                    else:
                        break
            time.sleep(self._loop_wait)
        logger.info(f"quitting run() of {self.__class__.__name__} of {self._target[0]}")
        self._release_lock()
        self._unregister_self()

    def deactivate(self, timeout: float = _DEFAULT_RELEASE_TIMEOUT) -> bool:
        # Notes: must be called from the other thread than one running self.
        self._to_release = True
        joined = False
        try:
            self.join(timeout)
            joined = True
        except TimeoutError:
            logger.warning(f"failed to join the lock keeper of {self._target[0]}")

        if self.has_lock:
            # Notes: this should not happen if joined successfully, and join should work.
            logger.error(f"failed to unlock {self._target[0]} due to timeout")

        return joined


class DummyLockKeeper(AbstractLockKeeper):
    _DEFAULT_LOCK_EXPIRATION: float = 0  # Notes: dummy value

    def __init__(self, *, target: tuple[str, int], loop_wait: float = AbstractLockKeeper._DEFAULT_LOOP_WAIT):
        super().__init__(target=target, loop_wait=loop_wait)
        self._locked: bool = False  # Notes: state of dummy lock

    @property
    def has_lock(self) -> bool:
        return self._locked

    def _take_lock(self) -> bool:
        self._locked = True
        return True

    def _keep_lock(self) -> bool:
        return True

    def _release_lock(self) -> None:
        self._locked = False


class FileLockKeeper(AbstractLockKeeper):
    _DEFAULT_LOCK_EXPIRATION = 15.0  # [s]
    _FLOCK_TIMEOUT: int = 5  # [s]
    _KILL_CHECK_PERIOD: int = 10

    def __init__(
        self,
        *,
        target: tuple[str, int],
        loop_wait: float = AbstractLockKeeper._DEFAULT_LOOP_WAIT,
        lock_directory: Path = _DEFAULT_LOCK_DIRECTORY,
    ):
        super().__init__(target=target, loop_wait=loop_wait)
        if not lock_directory.is_dir():
            raise RuntimeError(f"lock directory '{lock_directory}' is unavailable")
        self._lockfile: str = str(lock_directory / self._target[0])
        self._lockobj: fl.Lock = fl.Lock(lockfile=self._lockfile, lifetime=int(self._DEFAULT_LOCK_EXPIRATION))

        self._killfile: str = f"{self._lockobj.claimfile}" + ".kill"
        self._kill_check_counter = cycle([True] + (self._KILL_CHECK_PERIOD - 1) * [False])

    @property
    def has_lock(self) -> bool:
        try:
            return self._lockobj.is_locked
        except FileNotFoundError as e:
            logger.warning(e)
            return False

    def _take_lock(self) -> bool:
        try:
            self._lockobj.lock(timeout=self._FLOCK_TIMEOUT)
            logger.info(f"successfully acquired the lock of {self._target[0]}")
            return True
        except fl.TimeOutError:
            logger.warning(f"failed to acquire lock of {self._target[0]}")
            return False

    def _keep_lock(self) -> bool:
        if next(self._kill_check_counter):
            if self._check_if_killfile_exists():
                logger.info(f"Cancelled to refresh lock of {self._target[0]} due to the killfile.")
                return False

        try:
            self._lockobj.refresh()
        except Exception as e:
            logger.warning(e)
            return False
        return True

    def _check_if_killfile_exists(self) -> bool:
        return os.path.exists(self._killfile)

    def _release_lock(self) -> None:
        try:
            self._lockobj.unlock()
            logger.info(f"lock of {self._target[0]} is released successfully")
        except fl.NotLockedError:
            pass


atexit.register(AbstractLockKeeper.release_lock_all)


class _ExstickgeSockClientBase(_ExstickgeProxyBase):
    DEFAULT_RESPONSE_TIMEOUT: float = 0.5
    DEFAULT_PORT: int = 16384

    _PACKET_FORMAT = "!BBLH"  # MODE, I/F, ADDR, VALUE

    _MODE_READ_CMD = 0x80
    _MODE_READ_RPL = 0x81
    _MODE_WRITE_CMD = 0x82
    _MODE_WRITE_RPL = 0x83

    _MAX_RECV_TRIAL = 1000000  # for safety, must be harmless under the bombardment of wrong packtes.

    _SPIIF_MAPPINGS: Mapping[LsiKindId, Mapping[int, Tuple[int, int]]] = {}

    def __init__(
        self,
        target_address: str,
        target_port: int = DEFAULT_PORT,
        timeout: float = DEFAULT_RESPONSE_TIMEOUT,
        receiver_limit_by_binding: bool = False,
    ):
        super().__init__(target_address, target_port, timeout)
        self._receiver_limit_by_binding: bool = receiver_limit_by_binding
        self._sock: Union[socket.socket, None] = None
        self._lock: threading.Lock = threading.Lock()
        # TODO: consider the best timing of lock keeper creation.
        self._lock_keeper: Union[AbstractLockKeeper, None] = self._create_lockkeeper()

    def initialize(self):
        with self._lock:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self._sock.settimeout(self._timeout)
            if self._receiver_limit_by_binding:
                self._sock.bind((self._get_my_ip_addr(), 0))
            else:
                self._sock.bind(("", 0))

    @abstractmethod
    def _create_lockkeeper(self) -> AbstractLockKeeper: ...

    @property
    def _socket(self) -> socket.socket:
        if self._sock is None:
            raise RuntimeError(f"proxy of {self._target[0]} is not initialized")
        return self._sock

    @property
    def has_lock(self) -> bool:
        return (self._lock_keeper is not None) and self._lock_keeper.has_lock

    def _get_my_ip_addr(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((self._target[0], 0))
        my_ip_addr = sock.getsockname()[0]
        sock.close()
        return my_ip_addr

    def _xlate(self, kind: LsiKindId, idx: int, addr: int) -> Tuple[int, int]:
        """Translating a triplet of lsi_type, lsi_idx, and register address into two-byte packet field.
        :param kind: a type of LSI.
        :param idx: an index of the LSI in the specified type.
        :param addr: an address of a register of the LSI to be accessed.
        :return: an encoded address word of a command packet.
        """

        if not LsiKindId.has_value(kind):
            raise ValueError(f"invalid Lsi identifier '{kind}'")

        if kind not in self._SPIIF_MAPPINGS:
            raise ValueError(f"no instance of '{kind}' is no available on this board")

        spiif_idx, cs_idx = self._SPIIF_MAPPINGS[kind][idx]

        if kind == LsiKindId.ADRF6780:
            return spiif_idx, (cs_idx << 6) | (addr & self._ADDR_MASKS[kind])
        elif kind == LsiKindId.LMX2594:
            return spiif_idx, (cs_idx << 7) | (addr & self._ADDR_MASKS[kind])
        elif kind == LsiKindId.AD5328:
            return spiif_idx, (cs_idx << 4) | (addr & self._ADDR_MASKS[kind])
        elif kind == LsiKindId.GPIO:
            return spiif_idx, cs_idx | (addr & self._ADDR_MASKS[kind])
        elif kind == LsiKindId.AD9082:
            return spiif_idx, (idx << 15) | (addr & self._ADDR_MASKS[kind])
        else:
            raise ValueError(f"unexpected LsiKindId for socket client: {kind.name}")

    def _make_pkt(self, mode, if_idx, addr, value) -> bytes:
        return struct.pack(self._PACKET_FORMAT, mode, if_idx, addr, value)

    def _make_readpkt(self, kind: LsiKindId, idx: int, addr: int) -> bytes:
        if_idx, csel_addr = self._xlate(kind, idx, addr)
        return self._make_pkt(self._MODE_READ_CMD, if_idx, csel_addr, 0)

    def _make_writepkt(self, kind: LsiKindId, idx: int, addr: int, value: int) -> bytes:
        if_idx, csel_addr = self._xlate(kind, idx, addr)
        return self._make_pkt(self._MODE_WRITE_CMD, if_idx, csel_addr, value)

    def _send_and_recv(self, cmd: bytes) -> Union[bytes, None]:
        # Notes: lock should be acquired by caller
        try:
            if self._dump_enable:
                cmdrepr = " ".join([f"{x:02x}" for x in cmd])
                logger.info(f"-> {cmdrepr} to {self._target}")

            time0 = time.time()
            self._socket.sendto(cmd, self._target)
            for _ in range(self._MAX_RECV_TRIAL):
                rpl, addr = self._socket.recvfrom(16)
                if self._dump_enable:
                    rplrepr = " ".join([f"{x:02x}" for x in rpl])
                    logger.info(f"-> {rplrepr} from {addr}")

                if len(rpl) == 8 and addr == self._target and rpl[0] == cmd[0] + 1 and rpl[1:6] == cmd[1:6]:
                    return rpl
                else:
                    rplrepr = " ".join([f"{x:02x}" for x in rpl])
                    if len(rplrepr) > 26:
                        rplrepr = rplrepr[:26] + "..."
                    logger.warning(f"unexpected packet '{rplrepr:s}' is read from {addr}, ignore it.")

                if time.time() - time0 >= self._timeout:
                    logger.error(f"failed to receive a packet from {self._target} due to time out.")
                    break
            else:
                logger.error("too much unexpected packets, give up to get the reply")
        except socket.timeout:
            logger.error(f"failed to send/receive a packet to/from {self._target} due to time out.")

        return None

    def read_reg(self, kind, idx, addr) -> Union[int, None]:
        with self._lock:
            # Notes: for CSS firmware corresponding to the sock client, any simultaneous accesses can fail.
            if not self.has_lock:
                raise BoxLockError(f"lock of {self._target[0]} is not available")
            cmd = self._make_readpkt(kind, idx, addr)
            rpl = self._send_and_recv(cmd)
            if rpl:
                return struct.unpack("!H", rpl[6:8])[0]
        return None

    def write_reg(self, kind, idx, addr, value) -> bool:
        with self._lock:
            if not self.has_lock:
                raise BoxLockError(f"lock of {self._target[0]} is not available")
            cmd = self._make_writepkt(kind, idx, addr, value)
            rpl = self._send_and_recv(cmd)
            if rpl:
                return struct.unpack("!H", rpl[6:8])[0] == value
        return False

    def terminate(self):
        # Notes: terminate() is call by __del__() of Quel1ConfigSubsystemRoot and __del__() of self.__del__() defined
        #        at _ExstickgeProxyBase. terminate() should be defined to be idempotent.
        if self._sock is not None:
            self._sock.close()
            self._sock = None

        if hasattr(self, "_lock_keeper") and self._lock_keeper is not None and self._lock_keeper.is_alive():
            if self._lock_keeper.deactivate():
                self._lock_keeper = None



import logging
from dataclasses import dataclass, field
from typing import Dict, Final, Union

from quel_ic_config.abstract_ic import (
    AbstractIcConfigHelper,
    AbstractIcMixin,
    AbstractIcReg,
    Gpio2,
    Gpio4,
    Gpio8,
    Gpio16,
    Uint16,
    b_1bf_bool,
    b_nbf,
    p_1bf_bool,
    p_nbf,
)

logger = logging.getLogger(__name__)


@dataclass
class PowerboardPwmFanMaster(AbstractIcReg):
    en: bool = field(default=True)

    def parse(self, v: int) -> None:
        self.en = p_1bf_bool(v, 0)

    def build(self) -> int:
        return b_1bf_bool(self.en, 0)


@dataclass
class PowerboardPwmHeaterMaster(AbstractIcReg):
    en: bool = field(default=False)

    def parse(self, v: int) -> None:
        self.en = p_1bf_bool(v, 0)

    def build(self) -> int:
        return b_1bf_bool(self.en, 0)


@dataclass
class PowerboardPwmFanHighCount(AbstractIcReg):
    h_count: int = field(default=499)

    def parse(self, v: int) -> None:
        self.h_count = p_nbf(v, 9, 0)

    def build(self) -> int:
        return b_nbf(self.h_count, 9, 0)


@dataclass
class PowerboardPwmHeaterHighCount(AbstractIcReg):
    h_count: int = field(default=0)

    def parse(self, v: int) -> None:
        self.h_count = p_nbf(v, 9, 0)

    def build(self) -> int:
        return b_nbf(self.h_count, 9, 0)


@dataclass
class PowerboardPwmHeaterDelayCount(AbstractIcReg):
    d_count: int = field(default=0)

    def parse(self, v: int) -> None:
        self.d_count = p_nbf(v, 9, 0)

    def build(self) -> int:
        return b_nbf(self.d_count, 9, 0)


PowerboardPwmRegs: Dict[int, type] = {
    # Fan
    0x00: PowerboardPwmFanMaster,
    0x01: Gpio2,
    0x04: PowerboardPwmFanHighCount,
    0x05: PowerboardPwmFanHighCount,
    # Heater
    0x08: PowerboardPwmHeaterMaster,
    0x09: Gpio16,
    0x0A: Gpio16,
    0x0B: Gpio8,
    0x0C: Uint16,
    0x0D: Uint16,
    0x0E: PowerboardPwmHeaterHighCount,
    0x0F: PowerboardPwmHeaterHighCount,
    0x10: PowerboardPwmHeaterHighCount,
    0x11: PowerboardPwmHeaterHighCount,
    0x12: PowerboardPwmHeaterHighCount,
    0x13: PowerboardPwmHeaterHighCount,
    0x14: PowerboardPwmHeaterHighCount,
    0x15: PowerboardPwmHeaterHighCount,
    0x16: PowerboardPwmHeaterHighCount,
    0x17: PowerboardPwmHeaterHighCount,
    0x18: PowerboardPwmHeaterHighCount,
    0x19: PowerboardPwmHeaterHighCount,
    0x1A: PowerboardPwmHeaterHighCount,
    0x1B: PowerboardPwmHeaterHighCount,
    0x1C: PowerboardPwmHeaterHighCount,
    0x1D: PowerboardPwmHeaterHighCount,
    0x1E: PowerboardPwmHeaterHighCount,
    0x1F: PowerboardPwmHeaterHighCount,
    0x20: PowerboardPwmHeaterHighCount,
    0x21: PowerboardPwmHeaterHighCount,
    0x22: PowerboardPwmHeaterHighCount,
    0x23: PowerboardPwmHeaterHighCount,
    0x24: PowerboardPwmHeaterHighCount,
    0x25: PowerboardPwmHeaterHighCount,
    0x26: PowerboardPwmHeaterHighCount,
    0x27: PowerboardPwmHeaterHighCount,
    0x28: PowerboardPwmHeaterHighCount,
    0x29: PowerboardPwmHeaterHighCount,
    0x2A: PowerboardPwmHeaterHighCount,
    0x2B: PowerboardPwmHeaterHighCount,
    0x2C: PowerboardPwmHeaterHighCount,
    0x2D: PowerboardPwmHeaterHighCount,
    0x2E: PowerboardPwmHeaterHighCount,
    0x2F: PowerboardPwmHeaterHighCount,
    0x30: PowerboardPwmHeaterHighCount,
    0x31: PowerboardPwmHeaterHighCount,
    0x32: PowerboardPwmHeaterHighCount,
    0x33: PowerboardPwmHeaterHighCount,
    0x34: PowerboardPwmHeaterHighCount,
    0x35: PowerboardPwmHeaterHighCount,
    0x36: PowerboardPwmHeaterDelayCount,
    0x37: PowerboardPwmHeaterDelayCount,
    0x38: PowerboardPwmHeaterDelayCount,
    0x39: PowerboardPwmHeaterDelayCount,
    0x3A: PowerboardPwmHeaterDelayCount,
    0x3B: PowerboardPwmHeaterDelayCount,
    0x3C: PowerboardPwmHeaterDelayCount,
    0x3D: PowerboardPwmHeaterDelayCount,
    0x3E: PowerboardPwmHeaterDelayCount,
    0x3F: PowerboardPwmHeaterDelayCount,
    0x40: PowerboardPwmHeaterDelayCount,
    0x41: PowerboardPwmHeaterDelayCount,
    0x42: PowerboardPwmHeaterDelayCount,
    0x43: PowerboardPwmHeaterDelayCount,
    0x44: PowerboardPwmHeaterDelayCount,
    0x45: PowerboardPwmHeaterDelayCount,
    0x46: PowerboardPwmHeaterDelayCount,
    0x47: PowerboardPwmHeaterDelayCount,
    0x48: PowerboardPwmHeaterDelayCount,
    0x49: PowerboardPwmHeaterDelayCount,
    0x4A: PowerboardPwmHeaterDelayCount,
    0x4B: PowerboardPwmHeaterDelayCount,
    0x4C: PowerboardPwmHeaterDelayCount,
    0x4D: PowerboardPwmHeaterDelayCount,
    0x4E: PowerboardPwmHeaterDelayCount,
    0x4F: PowerboardPwmHeaterDelayCount,
    0x50: PowerboardPwmHeaterDelayCount,
    0x51: PowerboardPwmHeaterDelayCount,
    0x52: PowerboardPwmHeaterDelayCount,
    0x53: PowerboardPwmHeaterDelayCount,
    0x54: PowerboardPwmHeaterDelayCount,
    0x55: PowerboardPwmHeaterDelayCount,
    0x56: PowerboardPwmHeaterDelayCount,
    0x57: PowerboardPwmHeaterDelayCount,
    0x58: PowerboardPwmHeaterDelayCount,
    0x59: PowerboardPwmHeaterDelayCount,
    0x5A: PowerboardPwmHeaterDelayCount,
    0x5B: PowerboardPwmHeaterDelayCount,
    0x5C: PowerboardPwmHeaterDelayCount,
    0x5D: PowerboardPwmHeaterDelayCount,
    # External Switch
    0x5E: Gpio4,
}


PowerboardPwmRegsName: Dict[str, int] = {
    # Fan
    "FanMaster": 0x00,
    "FanEnable": 0x01,
    "FanHighCount_0": 0x04,
    "FanHighCount_1": 0x05,
    # Heater,
    "HeaterMaster": 0x08,
    "HeaterEnable_L16": 0x09,
    "HeaterEnable_M16": 0x0A,
    "HeaterEnable_H8": 0x0B,
    "HeaterCycleCount_L": 0x0C,
    "HeaterCycleCount_H": 0x0D,
    "HeaterHighCount_00": 0x0E,
    "HeaterHighCount_01": 0x0F,
    "HeaterHighCount_02": 0x10,
    "HeaterHighCount_03": 0x11,
    "HeaterHighCount_04": 0x12,
    "HeaterHighCount_05": 0x13,
    "HeaterHighCount_06": 0x14,
    "HeaterHighCount_07": 0x15,
    "HeaterHighCount_08": 0x16,
    "HeaterHighCount_09": 0x17,
    "HeaterHighCount_10": 0x18,
    "HeaterHighCount_11": 0x19,
    "HeaterHighCount_12": 0x1A,
    "HeaterHighCount_13": 0x1B,
    "HeaterHighCount_14": 0x1C,
    "HeaterHighCount_15": 0x1D,
    "HeaterHighCount_16": 0x1E,
    "HeaterHighCount_17": 0x1F,
    "HeaterHighCount_18": 0x20,
    "HeaterHighCount_19": 0x21,
    "HeaterHighCount_20": 0x22,
    "HeaterHighCount_21": 0x23,
    "HeaterHighCount_22": 0x24,
    "HeaterHighCount_23": 0x25,
    "HeaterHighCount_24": 0x26,
    "HeaterHighCount_25": 0x27,
    "HeaterHighCount_26": 0x28,
    "HeaterHighCount_27": 0x29,
    "HeaterHighCount_28": 0x2A,
    "HeaterHighCount_29": 0x2B,
    "HeaterHighCount_30": 0x2C,
    "HeaterHighCount_31": 0x2D,
    "HeaterHighCount_32": 0x2E,
    "HeaterHighCount_33": 0x2F,
    "HeaterHighCount_34": 0x30,
    "HeaterHighCount_35": 0x31,
    "HeaterHighCount_36": 0x32,
    "HeaterHighCount_37": 0x33,
    "HeaterHighCount_38": 0x34,
    "HeaterHighCount_39": 0x35,
    "HeaterDelayCount_00": 0x36,
    "HeaterDelayCount_01": 0x37,
    "HeaterDelayCount_02": 0x38,
    "HeaterDelayCount_03": 0x39,
    "HeaterDelayCount_04": 0x3A,
    "HeaterDelayCount_05": 0x3B,
    "HeaterDelayCount_06": 0x3C,
    "HeaterDelayCount_07": 0x3D,
    "HeaterDelayCount_08": 0x3E,
    "HeaterDelayCount_09": 0x3F,
    "HeaterDelayCount_10": 0x40,
    "HeaterDelayCount_11": 0x41,
    "HeaterDelayCount_12": 0x42,
    "HeaterDelayCount_13": 0x43,
    "HeaterDelayCount_14": 0x44,
    "HeaterDelayCount_15": 0x45,
    "HeaterDelayCount_16": 0x46,
    "HeaterDelayCount_17": 0x47,
    "HeaterDelayCount_18": 0x48,
    "HeaterDelayCount_19": 0x49,
    "HeaterDelayCount_20": 0x4A,
    "HeaterDelayCount_21": 0x4B,
    "HeaterDelayCount_22": 0x4C,
    "HeaterDelayCount_23": 0x4D,
    "HeaterDelayCount_24": 0x4E,
    "HeaterDelayCount_25": 0x4F,
    "HeaterDelayCount_26": 0x50,
    "HeaterDelayCount_27": 0x51,
    "HeaterDelayCount_28": 0x52,
    "HeaterDelayCount_29": 0x53,
    "HeaterDelayCount_30": 0x54,
    "HeaterDelayCount_31": 0x55,
    "HeaterDelayCount_32": 0x56,
    "HeaterDelayCount_33": 0x57,
    "HeaterDelayCount_34": 0x58,
    "HeaterDelayCount_35": 0x59,
    "HeaterDelayCount_36": 0x5A,
    "HeaterDelayCount_37": 0x5B,
    "HeaterDelayCount_38": 0x5C,
    "HeaterDelayCount_39": 0x5D,
    # External Switch
    "ExtSw": 0x5E,
}


class PowerboardPwmMixin(AbstractIcMixin):
    Regs = PowerboardPwmRegs
    RegNames = PowerboardPwmRegsName

    NUM_HEATER: Final[int] = 40

    def __init__(self, name: str):
        super().__init__(name)

    def dump_regs(self) -> Dict[int, int]:
        """dump all the available registers.
        :return: a mapping between an address and a value of the registers
        """
        regs = {}
        for addr in self.Regs:
            regs[addr] = self.read_reg(addr)
        return regs

    def _validate_fan_idx(self, idx: int) -> None:
        if idx not in {0, 1}:
            raise ValueError(f"illegal index of fan: {idx}")

    def get_fan_speed(self, idx: int) -> float:
        self._validate_fan_idx(idx)
        r = PowerboardPwmFanHighCount()
        r.parse(self.read_reg(self.RegNames[f"FanHighCount_{idx}"]))
        return r.h_count / 1000.0

    def set_fan_speed(self, idx: int, ratio: float) -> None:
        self._validate_fan_idx(idx)
        if not 0.0 <= ratio <= 1.0:
            clipped_ratio = max(0.0, min(ratio, 1.0))
            logger.warning(f"illegal ratio of fan[{idx}]: {ratio:.3f}, clipped to {clipped_ratio:.3f}")
            ratio = clipped_ratio
        self.write_reg(
            self.RegNames[f"FanHighCount_{idx}"], PowerboardPwmFanHighCount(h_count=round(ratio * 1000)).build()
        )

    def _validate_heater_idx(self, idx: int) -> None:
        if not isinstance(idx, int):
            raise TypeError(f"unexpected type of value for heater index: '{idx}'")
        if not (0 <= idx < self.NUM_HEATER):
            raise ValueError(f"illegal index of heater index: {idx}")

    def get_heater_settings(self, idx: int) -> Dict[str, float]:
        self._validate_heater_idx(idx)
        hr = PowerboardPwmHeaterHighCount()
        hr.parse(self.read_reg(self.RegNames[f"HeaterHighCount_{idx:02}"]))
        dr = PowerboardPwmHeaterDelayCount()
        dr.parse(self.read_reg(self.RegNames[f"HeaterDelayCount_{idx:02}"]))
        return {"high_ratio": hr.h_count / 1000.0, "delay_ratio": dr.d_count / 1000.0}

    def set_heater_settings(
        self, idx: int, *, high_ratio: Union[float, None] = None, delay_ratio: Union[float, None] = None
    ) -> None:
        self._validate_heater_idx(idx)
        if high_ratio is not None:
            if not 0.0 <= high_ratio <= 1.0:
                clipped_high_count = max(0.0, min(high_ratio, 1.0))
                logger.warning(
                    f"illegal on-ratio of heater[{idx}]: {high_ratio:.3f}, clipped to {clipped_high_count:.3f}"
                )
                high_ratio = clipped_high_count
        if delay_ratio is not None:
            if not 0.0 <= delay_ratio <= 1.0:
                clipped_delay_count = max(0.0, min(delay_ratio, 1.0))
                logger.warning(
                    f"illegal delay-ratio of heater[{idx}]: {delay_ratio:.3f}, clipped to {clipped_delay_count:.3f}"
                )
                delay_ratio = clipped_delay_count

        if high_ratio is not None:
            hr = PowerboardPwmHeaterHighCount(h_count=round(high_ratio * 1000))
            self.write_reg(self.RegNames[f"HeaterHighCount_{idx:02}"], hr.build())
        if delay_ratio is not None:
            dr = PowerboardPwmHeaterDelayCount(d_count=round(delay_ratio * 1000))
            self.write_reg(self.RegNames[f"HeaterDelayCount_{idx:02}"], dr.build())

    def get_heater_enable(self, idx: int) -> bool:
        self._validate_heater_idx(idx)
        reg_addr = self.RegNames["HeaterEnable_L16"] + (idx // 16)
        bit_pos = idx % 16
        return (self.read_reg(reg_addr) >> bit_pos) & 0x01 == 0x01

    def set_heater_enable(self, idx: int, enable: bool) -> None:
        self._validate_heater_idx(idx)
        reg_addr = self.RegNames["HeaterEnable_L16"] + (idx // 16)
        bit_pos = idx % 16
        v = self.read_reg(reg_addr) & ~(1 << bit_pos)
        if enable:
            v |= 1 << bit_pos
        self.write_reg(reg_addr, v)

    def clear_heater_enable_all(self):
        reg_addr = self.RegNames["HeaterEnable_L16"]
        self.write_reg(reg_addr, 0)
        self.write_reg(reg_addr + 1, 0)
        self.write_reg(reg_addr + 2, 0)

    def dump_heater_enable_all(self):
        def reverse_bin(v: int) -> str:
            vl = list(f"{v:08b}")
            vl.reverse()
            return "".join(vl)

        reg_addr = self.RegNames["HeaterEnable_L16"]
        logger.info(f"heater 00--07: {reverse_bin(self.read_reg(reg_addr) & 0xff)}")
        logger.info(f"heater 08--15: {reverse_bin(self.read_reg(reg_addr) >> 8)}")
        logger.info(f"heater 16--23: {reverse_bin(self.read_reg(reg_addr+1) & 0xff)}")
        logger.info(f"heater 24--31: {reverse_bin(self.read_reg(reg_addr+1) >> 8)}")
        logger.info(f"heater 32--39: {reverse_bin(self.read_reg(reg_addr+2) & 0xff)}")

    def get_heater_master(self) -> bool:
        r = PowerboardPwmHeaterMaster()
        r.parse(self.read_reg(self.RegNames["HeaterMaster"]))
        return r.en

    def set_heater_master(self, en: bool) -> None:
        r = PowerboardPwmHeaterMaster()
        r.en = en
        self.write_reg(self.RegNames["HeaterMaster"], r.build())


class PowerboardPwmConfigHelper(AbstractIcConfigHelper):
    def __init__(self, ic: PowerboardPwmMixin):
        super().__init__(ic)

    def flush(self, discard_after_flush=True):
        for addr, value in self.updated.items():
            self.ic.write_reg(addr, value)
        if discard_after_flush:
            self.discard()



import logging
import time
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Dict, Final, List, Sequence, Tuple, Union, cast

from quel_ic_config.abstract_ic import (
    AbstractIcConfigHelper,
    AbstractIcMixin,
    AbstractIcReg,
    b_1bf_bool,
    b_nbf,
    p_1bf_bool,
    p_nbf,
)

logger = logging.getLogger(__name__)


# General Registers
@dataclass
class Lmx2594R0(AbstractIcReg):
    ramp_en: bool = field(default=False)  # [15]
    vco_phase_sync_en: bool = field(default=False)  # [14]
    out_mute: bool = field(default=False)  # [9]
    fcal_hpfd_adj: int = field(default=0)  # [8:7]
    fcal_lpfd_adj: int = field(default=0)  # [6:5]
    fcal_en: bool = field(default=False)  # [3]
    muxout_ld_sel: bool = field(default=False)  # [2]
    reset: bool = field(default=False)  # [1]
    powerdown: bool = field(default=False)  # [0]

    def parse(self, v):
        self.ramp_en = p_1bf_bool(v, 15)
        self.vco_phase_sync_en = p_1bf_bool(v, 14)
        self.out_mute = p_1bf_bool(v, 9)
        self.fcal_hpfd_adj = p_nbf(v, 8, 7)
        self.fcal_lpfd_adj = p_nbf(v, 6, 5)
        self.fcal_en = p_1bf_bool(v, 3)
        self.muxout_ld_sel = p_1bf_bool(v, 2)
        self.reset = p_1bf_bool(v, 1)
        self.powerdown = p_1bf_bool(v, 0)

    def build(self):
        return (
            0b0010010000010000
            | b_1bf_bool(self.ramp_en, 15)
            | b_1bf_bool(self.vco_phase_sync_en, 14)
            | b_1bf_bool(self.out_mute, 9)
            | b_nbf(self.fcal_hpfd_adj, 8, 7)
            | b_nbf(self.fcal_lpfd_adj, 6, 5)
            | b_1bf_bool(self.fcal_en, 3)
            | b_1bf_bool(self.muxout_ld_sel, 2)
            | b_1bf_bool(self.reset, 1)
            | b_1bf_bool(self.powerdown, 0)
        )


@dataclass
class Lmx2594R1(AbstractIcReg):
    cal_clk_div: int = field(default=0)  # [2:0], 3@reset

    def parse(self, v):
        self.cal_clk_div = p_nbf(v, 2, 0)

    def build(self):
        return 0b0000100000001000 | b_nbf(self.cal_clk_div, 2, 0)


@dataclass
class Lmx2594R7(AbstractIcReg):
    out_force: bool = field(default=False)  # [14]

    def parse(self, v):
        self.out_force = p_1bf_bool(v, 14)

    def build(self):
        return 0b0000000010110010 | b_1bf_bool(self.out_force, 14)


# Input Path Registers
@dataclass
class Lmx2594R9(AbstractIcReg):
    osc_2x: bool = field(default=False)  # [12]

    def parse(self, v):
        self.osc_2x = p_1bf_bool(v, 12)

    def build(self):
        return 0b0000011000000100 | b_1bf_bool(self.osc_2x, 12)


@dataclass
class Lmx2594R10(AbstractIcReg):
    mult: int = field(default=0)  # [11:7], 1@reset

    def parse(self, v):
        self.mult = p_nbf(v, 11, 7)

    def build(self):
        return 0b0001000001011000 | b_nbf(self.mult, 11, 7)


@dataclass
class Lmx2594R11(AbstractIcReg):
    pll_r: int = field(default=0)  # [11:4], 1@reset

    def parse(self, v):
        self.pll_r = p_nbf(v, 11, 4)

    def build(self):
        return 0b0000000000001000 | b_nbf(self.pll_r, 11, 4)


@dataclass
class Lmx2594R12(AbstractIcReg):
    pll_r_pre: int = field(default=0)  # [11:0], 1@reset

    def parse(self, v):
        self.pll_r_pre = p_nbf(v, 11, 0)

    def build(self):
        return 0b0101000000000000 | b_nbf(self.pll_r_pre, 11, 0)


# Charge Pump Registers
@dataclass
class Lmx2594R14(AbstractIcReg):
    cpg: int = field(default=0)  # [6:4], 7@reset

    def parse(self, v):
        self.cpg = p_nbf(v, 6, 4)

    def build(self):
        return 0b0001111000000000 | b_nbf(self.cpg, 6, 4)


# VCO calibration registers
@dataclass
class Lmx2594R4(AbstractIcReg):
    acal_cmp_dly: int = field(default=0)  # [15:8], 10@reset

    def parse(self, v):
        self.acal_cmp_dly = p_nbf(v, 15, 8)

    def build(self):
        return 0b0000000001000011 | b_nbf(self.acal_cmp_dly, 15, 8)


@dataclass
class Lmx2594R8(AbstractIcReg):
    vco_daciset_force: bool = field(default=False)  # [14]
    vco_capctrl_force: bool = field(default=False)  # [11]

    def parse(self, v):
        self.vco_daciset_force = p_1bf_bool(v, 14)
        self.vco_capctrl_force = p_1bf_bool(v, 11)

    def build(self):
        return 0b0010000000000000 | b_1bf_bool(self.vco_daciset_force, 14) | b_1bf_bool(self.vco_capctrl_force, 11)


@dataclass
class Lmx2594R16(AbstractIcReg):
    vco_daciset: int = field(default=0)  # [8:0], 128@reset

    def parse(self, v):
        self.vco_daciset = p_nbf(v, 8, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.vco_daciset, 8, 0)


@dataclass
class Lmx2594R17(AbstractIcReg):
    vco_daciset_strt: int = field(default=0)  # [8:0]. 250@reset

    def parse(self, v):
        self.vco_daciset_strt = p_nbf(v, 8, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.vco_daciset_strt, 8, 0)


@dataclass
class Lmx2594R19(AbstractIcReg):
    vco_capctrl: int = field(default=0)  # [7:0], 183@reset

    def parse(self, v):
        self.vco_capctrl = p_nbf(v, 7, 0)

    def build(self):
        return 0b0010011100000000 | b_nbf(self.vco_capctrl, 7, 0)


@dataclass
class Lmx2594R20(AbstractIcReg):
    vco_sel: int = field(default=0)  # [13:11], 7@reset
    vco_sel_force: bool = field(default=False)  # [10]

    def parse(self, v):
        self.vco_sel = p_nbf(v, 13, 11)
        self.vco_sel_force = p_1bf_bool(v, 10)

    def build(self):
        return 0b1100000001001000 | b_nbf(self.vco_sel, 13, 11) | b_1bf_bool(self.vco_sel_force, 10)


@dataclass
class Lmx2594R78(AbstractIcReg):
    ramp_thresh_32: int = field(default=0)  # [11:11]
    quick_recal_en: bool = field(default=False)  # [9]
    vco_capctrl_strt: int = field(default=0)  # [8:1]

    def parse(self, v: int) -> None:
        self.ramp_thresh_32 = p_nbf(v, 11, 11)
        self.quick_recal_en = p_1bf_bool(v, 9)
        self.vco_capctrl_strt = p_nbf(v, 8, 1)

    def build(self) -> int:
        return (
            0b0000000000000001
            | b_nbf(self.ramp_thresh_32, 11, 11)
            | b_1bf_bool(self.quick_recal_en, 9)
            | b_nbf(self.vco_capctrl_strt, 8, 1)
        )


# N Divider, MASH, and Output Registers
@dataclass
class Lmx2594R34(AbstractIcReg):
    pll_n_18_16: int = field(default=0)  # [2:0]

    def parse(self, v):
        self.pll_n_18_16 = p_nbf(v, 2, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.pll_n_18_16, 2, 0)


@dataclass
class Lmx2594R36(AbstractIcReg):
    pll_n: int = field(default=0)  # [15:0], 100@reset

    def parse(self, v):
        self.pll_n = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.pll_n, 15, 0)


@dataclass
class Lmx2594R37(AbstractIcReg):
    mash_seed_en: bool = field(default=False)  # [15]
    pfd_dly_sel: int = field(default=0)  # [13:8]  2@reset

    def parse(self, v):
        self.mash_seed_en = p_1bf_bool(v, 15)
        self.pfd_dly_sel = p_nbf(v, 13, 8)

    def build(self):
        return 0b0000000000000100 | b_1bf_bool(self.mash_seed_en, 15) | b_nbf(self.pfd_dly_sel, 13, 8)


@dataclass
class Lmx2594R38(AbstractIcReg):
    pll_den_31_16: int = field(default=0)  # [15:0], 4294967295@reset, with R39

    def parse(self, v):
        self.pll_den_31_16 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.pll_den_31_16, 15, 0)


@dataclass
class Lmx2594R39(AbstractIcReg):
    pll_den_15_0: int = field(default=0)  # [15:0]

    def parse(self, v):
        self.pll_den_15_0 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.pll_den_15_0, 15, 0)


@dataclass
class Lmx2594R40(AbstractIcReg):
    mash_seed_31_16: int = field(default=0)  # [15:0]

    def parse(self, v):
        self.mash_seed_31_16 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.mash_seed_31_16, 15, 0)


@dataclass
class Lmx2594R41(AbstractIcReg):
    mash_seed_15_0: int = field(default=0)  # [15:0]

    def parse(self, v):
        self.mash_seed_15_0 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.mash_seed_15_0, 15, 0)


@dataclass
class Lmx2594R42(AbstractIcReg):
    pll_num_31_16: int = field(default=0)  # [15:0]

    def parse(self, v):
        self.pll_num_31_16 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.pll_num_31_16, 15, 0)


@dataclass
class Lmx2594R43(AbstractIcReg):
    pll_num_15_0: int = field(default=0)  # [15:0]

    def parse(self, v):
        self.pll_num_15_0 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.pll_num_15_0, 15, 0)


@dataclass
class Lmx2594R44(AbstractIcReg):
    outa_pwr: int = field(default=0)  # [13:8], 31@reset
    outb_pd: bool = field(default=False)  # [7], 1@reset
    outa_pd: bool = field(default=False)  # [6]
    mash_reset_n: bool = field(default=False)  # [5] 1@reset
    mash_order: int = field(default=0)  # [2:0]

    def parse(self, v):
        self.outa_pwr = p_nbf(v, 13, 8)
        self.outb_pd = p_1bf_bool(v, 7)
        self.outa_pd = p_1bf_bool(v, 6)
        self.mash_reset_n = p_1bf_bool(v, 5)
        self.mash_order = p_nbf(v, 2, 0)

    def build(self):
        return (
            0b0000000000000000
            | b_nbf(self.outa_pwr, 13, 8)
            | b_1bf_bool(self.outb_pd, 7)
            | b_1bf_bool(self.outa_pd, 6)
            | b_1bf_bool(self.mash_reset_n, 5)
            | b_nbf(self.mash_order, 2, 0)
        )


@dataclass
class Lmx2594R45(AbstractIcReg):
    outa_mux: int = field(default=0)  # [12:11], 1@reset
    out_iset: int = field(default=0)  # [10:9]
    outb_pwr: int = field(default=0)  # [5:0], 31@reset

    def parse(self, v):
        self.outa_mux = p_nbf(v, 12, 11)
        self.out_iset = p_nbf(v, 10, 9)
        self.outb_pwr = p_nbf(v, 5, 0)

    def build(self):
        return (
            0b1100000011000000 | b_nbf(self.outa_mux, 12, 11) | b_nbf(self.out_iset, 10, 9) | b_nbf(self.outb_pwr, 5, 0)
        )


@dataclass
class Lmx2594R46(AbstractIcReg):
    outb_mux: int = field(default=0)  # [1:0], 1@reset

    def parse(self, v):
        self.outb_mux = p_nbf(v, 1, 0)

    def build(self):
        return 0b0000011111111100 | b_nbf(self.outb_mux, 1, 0)


# SYNC and SysRefReq Input Pin Register
@dataclass
class Lmx2594R58(AbstractIcReg):
    inpin_ignore: bool = field(default=False)  # [15], 1@reset
    inpin_hyst: bool = field(default=False)  # [14]
    inpin_lvl: int = field(default=0)  # [13:12]
    inpin_fmt: int = field(default=0)  # [11:9]

    def parse(self, v):
        self.inpin_ignore = p_1bf_bool(v, 15)
        self.inpin_hyst = p_1bf_bool(v, 14)
        self.inpin_lvl = p_nbf(v, 13, 12)
        self.inpin_fmt = p_nbf(v, 11, 9)

    def build(self):
        return (
            0b0000000000000001
            | b_1bf_bool(self.inpin_ignore, 15)
            | b_1bf_bool(self.inpin_hyst, 14)
            | b_nbf(self.inpin_lvl, 13, 12)
            | b_nbf(self.inpin_fmt, 11, 9)
        )


# Lock Detect Registers
@dataclass
class Lmx2594R59(AbstractIcReg):
    ld_type: bool = field(default=False)  # [0], 1@reset

    def parse(self, v):
        self.ld_type = p_1bf_bool(v, 0)

    def build(self):
        return 0b0000000000000000 | b_1bf_bool(self.ld_type, 0)


@dataclass
class Lmx2594R60(AbstractIcReg):
    ld_dly: int = field(default=0)  # [15:0], 1000@reset

    def parse(self, v):
        self.ld_dly = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.ld_dly, 15, 0)


# MASH_RESET
@dataclass
class Lmx2594R69(AbstractIcReg):
    mash_reset_count_31_16: int = field(default=0)  # [15:0], 50000@reset (with R70)

    def parse(self, v):
        self.mash_reset_count_31_16 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.mash_reset_count_31_16, 15, 0)


@dataclass
class Lmx2594R70(AbstractIcReg):
    mash_reset_count_15_0: int = field(default=0)  # [15:0], 50000@reset (with R69)

    def parse(self, v):
        self.mash_reset_count_15_0 = p_nbf(v, 15, 0)

    def build(self):
        return b_nbf(self.mash_reset_count_15_0, 15, 0)


# SysREF Registers
@dataclass
class Lmx2594R71(AbstractIcReg):
    sysref_div_pre: int = field(default=0)  # [7:5], 4@reset
    sysref_pulse: bool = field(default=False)  # [4]
    sysref_en: bool = field(default=False)  # [3]
    sysref_repeat: bool = field(default=False)  # [2]

    def parse(self, v):
        self.sysref_div_pre = p_nbf(v, 7, 5)
        self.sysref_pulse = p_1bf_bool(v, 4)
        self.sysref_en = p_1bf_bool(v, 3)
        self.sysref_repeat = p_1bf_bool(v, 2)

    def build(self):
        return (
            0b0000000000000001
            | b_nbf(self.sysref_div_pre, 7, 5)
            | b_1bf_bool(self.sysref_pulse, 4)
            | b_1bf_bool(self.sysref_en, 3)
            | b_1bf_bool(self.sysref_repeat, 2)
        )


@dataclass
class Lmx2594R72(AbstractIcReg):
    sysref_div: int = field(default=0)  # [10:0]

    def parse(self, v):
        self.sysref_div = p_nbf(v, 10, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.sysref_div, 10, 0)


@dataclass
class Lmx2594R73(AbstractIcReg):
    jesd_dac2_ctrl: int = field(default=0)  # [11:6]
    jesd_dac1_ctrl: int = field(default=0)  # [5:0], 63@reset

    def parse(self, v):
        self.jesd_dac2_ctrl = p_nbf(v, 11, 6)
        self.jesd_dac1_ctrl = p_nbf(v, 5, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.jesd_dac2_ctrl, 11, 6) | b_nbf(self.jesd_dac1_ctrl, 5, 0)


@dataclass
class Lmx2594R74(AbstractIcReg):
    sysref_pulse_cnt: int = field(default=0)  # [15:12]
    jesd_dac4_ctrl: int = field(default=0)  # [11:6]
    jesd_dac3_ctrl: int = field(default=0)  # [5:0]

    def parse(self, v):
        self.sysref_pulse_cnt = p_nbf(v, 15, 12)
        self.jesd_dac4_ctrl = p_nbf(v, 11, 6)
        self.jesd_dac3_ctrl = p_nbf(v, 5, 0)

    def build(self):
        return (
            0b0000000000000000
            | b_nbf(self.sysref_pulse_cnt, 15, 12)
            | b_nbf(self.jesd_dac4_ctrl, 11, 6)
            | b_nbf(self.jesd_dac3_ctrl, 5, 0)
        )


# CHANNEL Divider Registers
@dataclass
class Lmx2594R31(AbstractIcReg):
    chdiv_div2: bool = field(default=False)  # [14]

    def parse(self, v):
        self.chdiv_div2 = p_1bf_bool(v, 14)

    def build(self):
        return 0b0000001111101100 | b_1bf_bool(self.chdiv_div2, 14)


@dataclass
class Lmx2594R75(AbstractIcReg):
    chdiv: int = field(default=0)  # [10:6]

    def parse(self, v):
        self.chdiv = p_nbf(v, 10, 6)

    def build(self):
        return 0b0000100000000000 | b_nbf(self.chdiv, 10, 6)


# Ramping and Calibration Fields
# TODO: define R79 and R80

# Ramping Registers/Ramp Limits
# TODO: define R81, R82, R83, R84, R85, and R86

# Ramping Registers/Ramping Triggers, Burst Mode, and RAMP0_RST
# TODO define R96 and R97

# Ramping Registers/Ramping Configuration
# TODO: define R98, R99, R100, R101, R102, R103, R104, R105, and R106


# Readback Registers
@dataclass
class Lmx2594R110(AbstractIcReg):
    rb_ld_vtune: int = field(default=0)  # [10:9]
    rb_vco_sel: int = field(default=0)  # [7:5]

    def parse(self, v):
        self.rb_ld_vtune = p_nbf(v, 10, 9)
        self.rb_vco_sel = p_nbf(v, 7, 5)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.rb_ld_vtune, 10, 9) | b_nbf(self.rb_vco_sel, 7, 5)


@dataclass
class Lmx2594R111(AbstractIcReg):
    rb_vco_capctrl: int = field(default=0)  # [7:0], 183@reset

    def parse(self, v):
        self.rb_vco_capctrl = p_nbf(v, 7, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.rb_vco_capctrl, 7, 0)


@dataclass
class Lmx2594R112(AbstractIcReg):
    rb_vco_daciset: int = field(default=0)  # [8:0], 170@reset

    def parse(self, v):
        self.rb_vco_daciset = p_nbf(v, 8, 0)

    def build(self):
        return 0b0000000000000000 | b_nbf(self.rb_vco_daciset, 8, 0)


Lmx2594Regs: Dict[int, type] = {
    0: Lmx2594R0,
    1: Lmx2594R1,
    4: Lmx2594R4,
    7: Lmx2594R7,
    8: Lmx2594R8,
    9: Lmx2594R9,
    10: Lmx2594R10,
    11: Lmx2594R11,
    12: Lmx2594R12,
    14: Lmx2594R14,
    16: Lmx2594R16,
    17: Lmx2594R17,
    19: Lmx2594R19,
    20: Lmx2594R20,
    31: Lmx2594R31,
    34: Lmx2594R34,
    36: Lmx2594R36,
    37: Lmx2594R37,
    38: Lmx2594R38,
    39: Lmx2594R39,
    40: Lmx2594R40,
    41: Lmx2594R41,
    42: Lmx2594R42,
    43: Lmx2594R43,
    44: Lmx2594R44,
    45: Lmx2594R45,
    46: Lmx2594R46,
    58: Lmx2594R58,
    59: Lmx2594R59,
    60: Lmx2594R60,
    69: Lmx2594R69,
    70: Lmx2594R70,
    71: Lmx2594R71,
    72: Lmx2594R72,
    73: Lmx2594R73,
    74: Lmx2594R74,
    75: Lmx2594R75,
    78: Lmx2594R78,
    110: Lmx2594R110,
    111: Lmx2594R111,
    112: Lmx2594R112,
}


Lmx2594RegNames: Dict[str, int] = {f"R{k}": k for k in Lmx2594Regs}


class Lmx2594LockStatus(IntEnum):
    TooLow = 0
    Invalid = 1
    Locked = 2
    TooHigh = 3


class Lmx2594Mixin(AbstractIcMixin):
    Regs: Dict[int, type] = Lmx2594Regs
    RegNames: Dict[str, int] = Lmx2594RegNames
    MIN_FREQ_MULTIPLIER: Final[int] = 75
    MAX_FREQ_MULTIPLIER: Final[int] = 150
    CHDIV2RATIO: Final[Dict[int, int]] = {
        0: 2,
        1: 4,
        2: 6,
        3: 8,
        4: 12,
        5: 16,
        6: 24,
        7: 32,
        8: 48,
        9: 64,
        10: 72,
        11: 96,
        12: 128,
        13: 192,
        14: 256,
        15: 384,
        16: 512,
        17: 768,
    }

    def __init__(self, name):
        super().__init__(name)

    def dump_regs(self) -> Dict[int, int]:
        regs: Dict[int, int] = {}
        for addr in range(0, 113):
            regs[addr] = self.read_reg(addr)
        return regs

    def soft_reset(self) -> None:
        v0 = self.read_reg(0) & ~0x08  # clear calibration_enable once just for the ease of debug
        self.write_reg(0, v0 | 0x2)
        time.sleep(0.01)
        self.write_reg(0, v0 & ~0x2)

    def is_locked(self) -> Lmx2594LockStatus:
        v110 = self.read_reg(110)
        return Lmx2594LockStatus((v110 >> 9) & 0x3)

    def calibrate(self) -> bool:
        # ensure it takes more than 10msec after manipulating any registers.
        time.sleep(0.01)

        # writing 1 on bit3 of R0 activates calibration procedure once. No transition from 0 to 1 is required.
        v0 = self.read_reg(0)
        self.write_reg(0, v0 | 0x08)
        # checking the completion of calibration based on register value.
        t0 = time.perf_counter()
        for i in range(2):
            n = (100, 500)[i]  # 500 is required for the conventional settings of the refclks.
            for _ in range(n):
                time.sleep(0.001)
                v110 = self.read_reg(110)
                if (v110 >> 9) & 0x3 == 2:
                    logger.info(f"calibration of {self.name} is finished in {(time.perf_counter()-t0)*1000:.1f}ms")
                    return True
            else:
                # Maybe the second kick is required due to inappropriate initial values of VCO related settings.
                # TODO: investigate it by make the register settings correct.
                if i == 0:
                    logger.info("raise calibration flag again")
                    self.write_reg(0, v0 | 0x08)
        else:
            logger.warning(f"calibration of {self.name} is not finished within {(time.perf_counter()-t0)*1000:.1f}ms")
        return False

    def _validate_freq_multiplier(self, freq_multiplier: int):
        if not isinstance(freq_multiplier, int):
            raise TypeError(f"unexpected frequency multiplier: {freq_multiplier}")
        if not (self.MIN_FREQ_MULTIPLIER <= freq_multiplier <= self.MAX_FREQ_MULTIPLIER):
            raise TypeError(f"invalid frequency multiplier: {freq_multiplier}")

    def set_lo_multiplier(self, freq_multiplier: int) -> None:
        addr34, reg34 = cast(Tuple[int, Lmx2594R34], self._read_and_parse_reg("R34"))
        addr36, reg36 = cast(Tuple[int, Lmx2594R36], self._read_and_parse_reg("R36"))
        reg34.pll_n_18_16 = (freq_multiplier >> 16) & 0x0007
        reg36.pll_n = freq_multiplier & 0xFFFF
        self._build_and_write_reg(addr34, reg34)
        self._build_and_write_reg(addr36, reg36)
        return

    def get_lo_multiplier(self) -> int:
        _, reg34 = cast(Tuple[int, Lmx2594R34], self._read_and_parse_reg("R34"))
        _, reg36 = cast(Tuple[int, Lmx2594R36], self._read_and_parse_reg("R36"))
        return (reg34.pll_n_18_16 << 16) + (reg36.pll_n)

    def _parse_divide_ratio(
        self, out_mux: Sequence[int], pwdn: Sequence[bool], enable_seg1: bool, chdiv: int
    ) -> Tuple[int, int]:
        ratio: List[int] = [0, 0]
        for i in range(2):
            if pwdn[i]:
                continue

            if out_mux[i] == 0:
                if enable_seg1:
                    if chdiv == 0:
                        raise RuntimeError(
                            f"invalid combination of seg1_enable(={enable_seg1}) and chdiv(= {chdiv}) at {self.name}"
                        )
                    elif chdiv in self.CHDIV2RATIO:
                        ratio[i] = self.CHDIV2RATIO[chdiv]
                    else:
                        raise RuntimeError(f"invalid chdiv(= {chdiv}) at {self.name}")
                else:
                    if chdiv == 0:
                        ratio[i] = 2
                    else:
                        raise RuntimeError(
                            f"invalid combination of seg1_enable(={enable_seg1}) and chdiv(= {chdiv}) at {self.name}"
                        )
            elif out_mux[i] == 1:
                ratio[i] = 1
            elif out_mux[i] == 2:
                # SYSREF output, considered as ration = 0
                pass
            elif out_mux[i] == 3:
                # high_impedance output, equivalent to ratio = 0
                pass
            else:
                # Note: Sysref should not be selected for LO.
                raise AssertionError(f"unexpected mux of outpin:{i} of {self.name}")

        return ratio[0], ratio[1]

    def set_divider_ratio(self, ratio0: Union[int, None], ratio1: Union[int, None]) -> None:
        addr44, reg44 = cast(Tuple[int, Lmx2594R44], self._read_and_parse_reg("R44"))
        addr45, reg45 = cast(Tuple[int, Lmx2594R45], self._read_and_parse_reg("R45"))
        addr46, reg46 = cast(Tuple[int, Lmx2594R46], self._read_and_parse_reg("R46"))
        addr31, reg31 = cast(Tuple[int, Lmx2594R31], self._read_and_parse_reg("R31"))
        addr75, reg75 = cast(Tuple[int, Lmx2594R75], self._read_and_parse_reg("R75"))

        out_mux: Tuple[int, int] = reg45.outa_mux, reg46.outb_mux
        pwdn: Tuple[bool, bool] = reg44.outa_pd, reg44.outb_pd
        enable_seg1: bool = reg31.chdiv_div2
        chdiv: int = reg75.chdiv

        ratios = list(self._parse_divide_ratio(out_mux, pwdn, enable_seg1, chdiv))
        if ratio0 is not None:
            ratios[0] = ratio0
        if ratio1 is not None:
            ratios[1] = ratio1
        if not (ratios[0] in {0, 1} or ratios[1] in {0, 1} or ratios[0] == ratios[1]):
            raise ValueError(f"impossible combination of divide ratios ({ratios[0]}, {ratios[1]}) for {self.name}")

        for i, ratio in enumerate(ratios):
            if ratio == 1:
                if i == 0:
                    reg45.outa_mux = 1
                    reg44.outa_pd = False
                elif i == 1:
                    reg46.outb_mux = 1
                    reg44.outb_pd = False
                else:
                    raise AssertionError
            elif ratio == 0:
                if i == 0:
                    reg45.outa_mux = 3
                    reg44.outa_pd = True
                elif i == 1:
                    reg46.outb_mux = 3
                    reg44.outb_pd = True
                else:
                    raise AssertionError
            else:
                if i == 0:
                    reg45.outa_mux = 0
                    reg44.outa_pd = False
                elif i == 1:
                    reg46.outb_mux = 0
                    reg44.outb_pd = False
                else:
                    raise AssertionError

        for i, ratio in enumerate(ratios):
            if ratio in {0, 1}:
                continue
            else:
                for k, v in self.CHDIV2RATIO.items():
                    if v == ratio:
                        chdiv = k
                        break
                else:
                    raise ValueError(f"invalid frequency divider ratio {ratio} for outpin {i} of {self.name}")
                reg31.chdiv_div2 = ratio != 2
                reg75.chdiv = chdiv
                break

        self._build_and_write_reg(75, reg75)
        self._build_and_write_reg(31, reg31)
        self._build_and_write_reg(46, reg46)
        self._build_and_write_reg(45, reg45)
        self._build_and_write_reg(44, reg44)

    def get_divider_ratio(self) -> Tuple[int, int]:
        addr44, reg44 = cast(Tuple[int, Lmx2594R44], self._read_and_parse_reg("R44"))
        addr45, reg45 = cast(Tuple[int, Lmx2594R45], self._read_and_parse_reg("R45"))
        addr46, reg46 = cast(Tuple[int, Lmx2594R46], self._read_and_parse_reg("R46"))
        addr31, reg31 = cast(Tuple[int, Lmx2594R31], self._read_and_parse_reg("R31"))
        addr75, reg75 = cast(Tuple[int, Lmx2594R75], self._read_and_parse_reg("R75"))

        out_mux: Tuple[int, int] = reg45.outa_mux, reg46.outb_mux
        pwdn: Tuple[bool, bool] = reg44.outa_pd, reg44.outb_pd
        enable_seg1: bool = reg31.chdiv_div2
        chdiv: int = reg75.chdiv
        return self._parse_divide_ratio(out_mux, pwdn, enable_seg1, chdiv)

    def get_sync_enable(self) -> bool:
        addr0, reg0 = cast(Tuple[int, Lmx2594R0], self._read_and_parse_reg("R0"))
        return reg0.vco_phase_sync_en


class Lmx2594ConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming LMX2594 with convenient notations. It also provides caching capability that
    keep modifications on the registers to wirte them at once with flash_updated() in a right order.
    """

    _TopWritableAddress: int = 106

    def __init__(self, ic: Lmx2594Mixin):
        super().__init__(ic)

    def flush(self, discard_after_flush=True):
        for addr in sorted(self.updated, reverse=True):
            if addr <= self._TopWritableAddress:
                self.ic.write_reg(addr, self.updated[addr])
            else:
                logger.warning(f"failed attempt to write reg-{addr} is detected")
        if discard_after_flush:
            self.discard()

    # TODO: will be moved to super class later.
    def load_settings(self, regs: Dict[str, Dict[str, Union[int, str]]]):
        pass



from .common_arguments import (
    add_common_arguments,
    add_common_workaround_arguments,
    complete_ipaddrs,
)
from .modified_config import modified_config
from .plot_iqs import plot_iqs
from .simple_multibox_framework import (
    BoxPool,
    BoxSettingType,
    VportSettingType,
    VportSimpleParamtersSettingType,
    VportTypicalSettingType,
    calc_angle,
    find_chunks,
    single_schedule,
)

__all__ = (
    "add_common_arguments",
    "add_common_workaround_arguments",
    "complete_ipaddrs",
    "plot_iqs",
    "BoxPool",
    "BoxSettingType",
    "VportSettingType",
    "VportSimpleParamtersSettingType",
    "VportTypicalSettingType",
    "calc_angle",
    "find_chunks",
    "single_schedule",
    "modified_config",
)



import logging
from contextlib import contextmanager
from typing import Any

import quel_ic_config as qi

logger = logging.getLogger(__name__)


@contextmanager
def modified_config(box: qi.Quel1Box, config: dict[str, Any]):
    original = box.dump_box()
    logger.info(f"Temporarily modifying config for box '{box.name}': {config}")
    box.config_box(config)
    yield box
    box.config_box(original)
    logger.info(f"Config for {box.name} has been restored.")



import logging
from collections.abc import Collection, Mapping, Sequence
from typing import Any, Final, Union, cast

import numpy as np
import numpy.typing as npt
from e7awghal import AwgParam, CapIqDataReader, CapParam, CapSection, WaveChunk

from quel_ic_config import (
    AbstractStartAwgunitsTask,
    BoxStartCapunitsByTriggerTask,
    Quel1Box,
    Quel1PortType,
    QuelClockMasterV1,
)

logger = logging.getLogger(__name__)

BoxSettingType = Mapping[str, Any]
VportTypicalSettingType = Mapping[str, Any]
VportSimpleParamtersSettingType = Mapping[int, Mapping[str, Any]]
VportSettingType = Union[VportTypicalSettingType, VportSimpleParamtersSettingType]


class BoxPool:
    SYSREF_PERIOD: Final[int] = 2000
    # TODO: tried to find the best value, but the best value changes link-up by link-up. so, calibration is required.
    TIMING_OFFSET: Final[int] = 0
    DEFAULT_NUM_SYSREF_MEASUREMENTS: Final[int] = 100

    def __init__(
        self,
        cm_settings: Mapping[str, Any],
        box_settings: Mapping[str, Mapping[str, BoxSettingType]],
        cap_settings: Mapping[str, Mapping[str, VportSettingType]],
        gen_settings: Mapping[str, Mapping[str, VportSettingType]],
    ):
        self._cm_settings: Mapping[str, Any] = cm_settings
        self._box_settings: Mapping[str, Mapping[str, Any]] = box_settings
        self._cap_settings: Mapping[str, Mapping[str, VportSettingType]] = cap_settings
        self._gen_settings: Mapping[str, Mapping[str, VportSettingType]] = gen_settings

        self._clock_master: Union[QuelClockMasterV1, None] = None

        self._boxes: dict[str, Quel1Box] = {}
        self._channels: dict[str, tuple[str, Quel1PortType, int]] = {}  # Notes: genname -> (boxname, port, channel)
        self._runits: dict[str, tuple[str, Quel1PortType, set[int]]] = {}  # Notes: capname -> (boxname, port, runits)
        self._linkstatus: dict[str, bool] = {}

        self._estimated_timediff: dict[str, int] = {boxname: 0 for boxname in self._boxes}
        self._cap_sysref_time_offset: int = 0

    def initialize(
        self, recreate_box=False, reconnect: bool = True, config_css: bool = True, allow_resync: bool = True
    ):
        if recreate_box or len(self._boxes) == 0:
            self.create_box(self._box_settings)
        if len(self._cm_settings) > 0:
            self._clock_master = QuelClockMasterV1(**self._cm_settings)
        self.scan_link_status(reconnect=reconnect)
        self.reset_wss()
        self.config_channels(config_css, True)
        self.config_runits(config_css, True)
        sync_status = self.check_synchronization()
        if not sync_status:
            if allow_resync:
                self.resync()
                sync_status = self.check_synchronization()
        if not sync_status:
            raise RuntimeError("synchronization error")

    def create_box(self, settings: Mapping[str, Mapping[str, Any]]):
        for k, v in settings.items():
            self._boxes[k] = Quel1Box.create(**v)
            self._linkstatus[k] = False

    def check_synchronization(self) -> bool:
        if len(self._boxes) < 2:
            return True

        fsync_ok = True
        sysref_counters = {name: box.get_averaged_sysref_offset() for name, box in self._boxes.items()}
        diff = max(sysref_counters.values()) - min(sysref_counters.values())
        if diff > 1000:
            diff = 2000 - diff  # Notes: abs(diff - 2000)
        fsync_ok &= diff < 100
        logger.info(f"synchronization status: {fsync_ok}, max_delta = {diff}")
        return fsync_ok

    def scan_link_status(self, reconnect=False):
        for name, box in self._boxes.items():
            link_status: bool = True
            if reconnect:
                if not all(box.reconnect().values()):
                    if all(box.reconnect(ignore_crc_error_of_mxfe=box.css.get_all_groups()).values()):
                        logger.warning(f"crc error has been detected on MxFEs of {name}")
                    else:
                        logger.error(f"datalink between MxFE and FPGA of {name} is not working")
                        link_status = False
            else:
                if not all(box.link_status().values()):
                    if all(box.link_status(ignore_crc_error_of_mxfe=box.css.get_all_groups()).values()):
                        logger.warning(f"crc error has been detected on MxFEs of {name}")
                    else:
                        logger.error(f"datalink between MxFE and FPGA of {name} is not working")
                        link_status = False
            self._linkstatus[name] = link_status

    def reset_wss(self):
        for name, box in self._boxes.items():
            box.initialize_all_awgunits()
            box.initialize_all_capunits()

    def config_channels(self, config_css: bool, config_wss: bool):
        for k, v in self._gen_settings.items():
            if "create" not in v:
                raise ValueError("no 'create' parameters for wavegen '{k}'")

            c = cast(VportTypicalSettingType, v["create"])
            boxname: str = c["boxname"]
            port: Quel1PortType = c["port"]
            channel: int = c["channel"]
            self._channels[k] = (boxname, port, channel)

            if config_css and "config" in v:
                self._config_channel(k, **cast(VportTypicalSettingType, v["config"]))
            if config_wss and "cw_parameter" in v:
                self._load_awgparam_cw(k, **cast(VportTypicalSettingType, v["cw_parameter"]))

    def config_runits(self, config_css: bool, config_wss: bool):
        for k, v in self._cap_settings.items():
            if "create" not in v:
                raise ValueError("no 'create' parameters for wavegen '{k}'")

            cr = cast(VportTypicalSettingType, v["create"])
            boxname: str = cr["boxname"]
            port: Quel1PortType = cr["port"]
            runits: set[int] = set(cr["runits"])
            self._runits[k] = (boxname, port, runits)

            if config_css and "config" in v:
                self._config_runits(k, **cast(VportTypicalSettingType, v["config"]))
            if config_wss and "simple_parameters" in v:
                self._load_capparam_simple(
                    k,
                    cast(
                        Mapping[int, Mapping[str, Any]], cast(VportSimpleParamtersSettingType, v["simple_parameters"])
                    ),
                )

    def _config_channel(self, gname: str, *, fnco_freq: Union[float, None] = None, **kwargs) -> None:
        logger.info(f"reconfigureing {gname}")
        boxname, port, channel = self._channels[gname]
        box = self._boxes[boxname]

        box.config_port(port=port, **kwargs)
        if fnco_freq is not None:
            box.config_channel(port=port, channel=channel, fnco_freq=fnco_freq)

    def _load_awgparam_cw(
        self,
        gname: str,
        amplitude: float,
        num_wave_sample: int,
        num_repeats: tuple[int, int],
        num_wait_samples: tuple[int, int],
        null_chunk_sample: Union[int, None] = None,
    ) -> None:
        boxname, port, channel = self._channels[gname]
        box = self._boxes[boxname]

        if num_wave_sample % 64 != 0:
            raise ValueError("num_wave_sample must be multiples of 64")
        if num_wait_samples[0] % 16 != 0:
            raise ValueError("both num_wait_samples[0] must be multiples of 16")
        if num_wait_samples[1] % 4 != 0:
            raise ValueError("both num_wait_samples must be multiples of 4")
        if null_chunk_sample is not None and null_chunk_sample < 64:
            raise ValueError("null_chunk_sample must not be less than 64")
        if null_chunk_sample is not None and null_chunk_sample % 4 != 0:
            raise ValueError("null_chunk_sample must be multiples of 4")

        cw_iq = np.zeros(num_wave_sample, dtype=np.complex64)
        cw_iq[:] = amplitude * (1.0 + 0.0j)
        box.register_wavedata(port, channel, "glc_cw", cw_iq)

        ap = AwgParam(num_wait_word=num_wait_samples[0] // 4, num_repeat=num_repeats[0])
        if null_chunk_sample is not None:
            ap.chunks.append(WaveChunk(name_of_wavedata="null", num_blank_word=(null_chunk_sample - 64) // 4))
        ap.chunks.append(WaveChunk(name_of_wavedata="glc_cw", num_blank_word=num_wait_samples[1] // 4))
        box.config_channel(port, channel, awg_param=ap)

    def _config_runits(self, cname: str, *, fnco_freq: Union[Mapping[int, float], float, None] = None, **kwargs):
        logger.info(f"reconfigureing {cname}")
        boxname, port, runits = self._runits[cname]
        box = self._boxes[boxname]

        box.config_port(port=port, **kwargs)
        if isinstance(fnco_freq, Mapping):
            for runit in runits:
                box.config_runit(port=port, runit=runit, fnco_freq=fnco_freq[runit])
        elif fnco_freq is not None:
            for runit in runits:
                box.config_runit(port=port, runit=runit, fnco_freq=fnco_freq)
        else:
            raise TypeError(f"malformed fnco_freq: {fnco_freq}")

    @staticmethod
    def _make_cwparam_simple(
        *,
        num_delay_sample: int,
        num_integration_section: int,
        num_capture_samples: Sequence[int],
        num_blank_samples: Sequence[int],
    ) -> CapParam:
        if num_delay_sample % 4 != 0:
            raise ValueError(f"num_delay_sample (= {num_delay_sample} is not multiple of 4.")

        cp = CapParam(num_wait_word=num_delay_sample // 4, num_repeat=num_integration_section)
        for idx in range(len(num_capture_samples)):
            if num_capture_samples[idx] % 4 != 0:
                raise ValueError(f"num_capture_samples[{idx}] (= {num_capture_samples[idx]}) is not multiple of 4")
            if num_blank_samples[idx] % 4 != 0:
                raise ValueError(f"num_blank_samples[{idx}] (= {num_blank_samples[idx]}) is not multiple of 4")
            cp.sections.append(
                CapSection(
                    name="s0",
                    num_capture_word=num_capture_samples[idx] // 4,
                    num_blank_word=num_blank_samples[idx] // 4,
                )
            )
        return cp

    def _load_capparam_simple(self, cname: str, simple_parameters: VportSimpleParamtersSettingType) -> None:
        boxname, port, runits = self._runits[cname]
        box = self._boxes[boxname]

        for runit, v in simple_parameters.items():
            if runit not in runits:
                raise ValueError(f"an invalid runit: {runit}")
            box.config_runit(port=port, runit=runit, capture_param=self._make_cwparam_simple(**v))

    def resync(self):
        if self._clock_master:
            self._clock_master.sync_boxes()

    def get_boxes(self) -> dict[str, Quel1Box]:
        return dict(self._boxes)

    def get_box(self, name: str) -> Quel1Box:
        if name in self._boxes:
            box = self._boxes[name]
            return box
        else:
            raise ValueError(f"invalid name of box: '{name}'")

    def get_linkstatus(self, name: str) -> bool:
        if name in self._boxes:
            return self._linkstatus[name]
        else:
            raise ValueError(f"invalid name of box: '{name}'")

    def _mod_by_sysref(self, t: int) -> int:
        h = self.SYSREF_PERIOD // 2
        return (t + h) % self.SYSREF_PERIOD - h
        # return t % self.SYSREF_PERIOD

    def check_background_noise(self, runit_names: set[str]) -> dict[str, float]:
        bgnoise: dict[str, float] = {}

        runits_by_box: dict[str, set[tuple[Quel1PortType, int]]] = {}
        for ru in runit_names:
            boxname, port, runits = self._runits[ru]
            if boxname not in runits_by_box:
                runits_by_box[boxname] = set()
            for runit in runits:
                runits_by_box[boxname].add((port, runit))

        for boxname, fqrunits in runits_by_box.items():
            ctask = self._boxes[boxname].start_capture_now(fqrunits)
            rdr = ctask.result()
            for fqrunit in fqrunits:
                bg = np.abs(rdr[fqrunit].as_wave_dict()["s0"][0])
                bgavg = np.mean(bg)
                bgmax = np.max(bg)
                logger.info(
                    f"background noise of port-#{fqrunit[0]:02d} of {boxname}: {bgavg:.1f} (mean), {bgmax:.1f} (max)"
                )

                # TODO: revise this!
                for rn, (bn, p, us) in self._runits.items():
                    for u in us:
                        if bn == boxname and p == fqrunit[0] and u == fqrunit[1]:
                            bgnoise[rn] = bgmax
        return bgnoise

    def measure_timediff(self, cname: str, num_iters: int = DEFAULT_NUM_SYSREF_MEASUREMENTS) -> None:
        counter_at_sysref_clk: dict[str, int] = {boxname: 0 for boxname in self._boxes}

        for i in range(num_iters):
            for boxname, box in self._boxes.items():
                last_sysref_time = box.get_latest_sysref_timecounter()
                counter_at_sysref_clk[boxname] += self._mod_by_sysref(last_sysref_time)

        avg: dict[str, int] = {boxname: round(cntr / num_iters) for boxname, cntr in counter_at_sysref_clk.items()}
        # Notes: averaged timecounter of capturing box since self._runits[cname]: tuple[str, int, set[int]]
        self._cap_sysref_time_offset = avg[self._runits[cname][0]]
        self._estimated_timediff = {boxname: cntr - self._cap_sysref_time_offset for boxname, cntr in avg.items()}
        logger.info(f"estimated time difference: {self._estimated_timediff}")

    def start_at(
        self,
        runit_name: str,
        channel_names: Collection[str],
        min_time_offset: int,
        displacement: int = 0,
    ) -> tuple[int, dict[str, BoxStartCapunitsByTriggerTask], dict[str, AbstractStartAwgunitsTask]]:
        if len(channel_names) == 0:
            logger.warning("no pulse generator to activate")

        capture_tasks: dict[str, BoxStartCapunitsByTriggerTask] = {}
        wavegen_tasks: dict[str, AbstractStartAwgunitsTask] = {}

        channels_by_box: dict[str, set[tuple[Quel1PortType, int]]] = {}
        for ch in channel_names:
            boxname, port, channel = self._channels[ch]
            if boxname not in channels_by_box:
                channels_by_box[boxname] = set()
            channels_by_box[boxname].add((port, channel))

        runits_by_box: dict[str, set[tuple[Quel1PortType, int]]] = {}
        for ru in (runit_name,):  # Notes: for future extention
            boxname, port, runits = self._runits[ru]
            if boxname not in channels_by_box:
                raise RuntimeError(f"impossible to trigger the runit of box '{boxname}'")
            if boxname not in runits_by_box:
                runits_by_box[boxname] = set()
            for runit in runits:
                runits_by_box[boxname].add((port, runit))

        # Notes: showing timecounters of each box (just for information)
        for boxname, chs in channels_by_box.items():
            box = self._boxes[boxname]
            if len(chs) == 0:
                continue
            ct = box.get_current_timecounter()
            lst = box.get_latest_sysref_timecounter()
            logger.info(
                f"boxname: {boxname}, current time: {ct}, "
                f"sysref offset: {self._mod_by_sysref(lst)}, "
                f"have_capture: {boxname in runits_by_box}"
            )

        cpbox0 = list(runits_by_box.keys())[0]
        current_time = self._boxes[cpbox0].get_current_timecounter()  # Notes: monitoring box provides standard time.
        last_sysref_time = self._boxes[cpbox0].get_latest_sysref_timecounter()

        # Notes: checking the fluctuation of sysref trigger (just for information).
        fluctuation = self._mod_by_sysref(last_sysref_time - self._cap_sysref_time_offset)
        if abs(fluctuation) > 4:
            logger.warning(
                f"large fluctuation (= {fluctuation}) of sysref is detected from the previous timing measurement"
            )

        # Notes: time adjustment among the boxes
        base_time = current_time + min_time_offset
        offset = (16 - (base_time - self._cap_sysref_time_offset) % 16) % 16
        base_time += offset
        base_time += displacement  # inducing clock displacement for performance evaluation (must be 0 usually).
        base_time += self.TIMING_OFFSET  # Notes: the safest timing to issue trigger, at the middle of two AWG block.

        # Notes: scheduling!
        for boxname in channels_by_box:
            ts = base_time + self._estimated_timediff[boxname]
            if boxname in runits_by_box:
                ctask, gtask = self._boxes[boxname].start_capture_by_awg_trigger(
                    runits_by_box[boxname], channels_by_box[boxname], timecounter=ts
                )
                capture_tasks[boxname] = ctask
                wavegen_tasks[boxname] = gtask
                logger.info(f"reserving capture and wavegen of {boxname} at {ts}")
            else:
                gtask = self._boxes[boxname].start_wavegen(channels_by_box[boxname], timecounter=ts)
                wavegen_tasks[boxname] = gtask
                logger.info(f"reserving wavegen of {boxname} at {ts}")
        logger.info("scheduling completed")

        return base_time, capture_tasks, wavegen_tasks


def single_schedule(
    cp: str, pgs: Collection[str], boxpool: BoxPool, power_thr: float, displacement: int = 0
) -> tuple[int, npt.NDArray[np.complex64], tuple[tuple[int, int], ...]]:
    base_time, c_tasks, g_tasks = boxpool.start_at(
        runit_name=cp, channel_names=pgs, min_time_offset=125_000_000 // 10, displacement=displacement
    )
    for boxname, g_task in g_tasks.items():
        g_task.result()
        logger.info(f"wave generation of box {boxname} is completed")

    rdrs: dict[str, dict[tuple[Quel1PortType, int], CapIqDataReader]] = {}
    for boxname, c_task in c_tasks.items():
        rdrs[boxname] = c_task.result()
        logger.info(f"capture of box {boxname} is completed")

    if len(rdrs) != 1:
        raise AssertionError("too much reader objects...")

    cp_box, cp_port, cp_runits = boxpool._runits[cp]

    rdr = rdrs[cp_box]
    iq = rdr[cp_port, list(cp_runits)[0]].as_wave_dict()["s0"][0]
    chunks = find_chunks(iq, power_thr=power_thr)
    return base_time, iq, chunks


def find_chunks(
    iq: npt.NDArray[np.complex64], power_thr=1000.0, space_thr=16, minimal_length=16
) -> tuple[tuple[int, int], ...]:
    chunk = (abs(iq) > power_thr).nonzero()[0]
    if len(chunk) == 0:
        logger.info("no pulse!")
        return ()

    gaps = (chunk[1:] - chunk[:-1]) > space_thr
    start_edges = list(chunk[1:][gaps])
    start_edges.insert(0, chunk[0])
    last_edges = list(chunk[:-1][gaps])
    last_edges.append(chunk[-1])
    chunks = tuple([(s, e) for s, e in zip(start_edges, last_edges) if e - s >= minimal_length])

    n_chunks = len(chunks)
    logger.info(f"number_of_chunks: {n_chunks}")
    for i, chunk in enumerate(chunks):
        s, e = chunk
        iq0 = np.average(iq[s:e])
        angle = round(np.arctan2(iq0.real, iq0.imag) * 180.0 / np.pi, 1)
        logger.info(f"  chunk {i}: {e - s} samples, ({s} -- {e}),  mean phase = {angle:.1f}")
    return chunks


def calc_angle(iq) -> tuple[float, float, float]:
    angle = np.angle(iq)
    min_angle = min(angle)
    max_angle = max(angle)
    if max_angle - min_angle > 6.0:
        angle = (angle + 2 * np.pi) % np.pi

    avg = np.mean(angle) * 180.0 / np.pi
    sd = np.sqrt(np.var(angle)) * 180.0 / np.pi
    delta = (max(angle) - min(angle)) * 180.0 / np.pi
    return avg, sd, delta



import ipaddress
import logging
import time
from abc import ABCMeta, abstractmethod
from collections.abc import Collection
from concurrent.futures import CancelledError, Future, ThreadPoolExecutor
from enum import Enum
from functools import cached_property
from threading import RLock
from typing import Callable, Final, Generic, Optional, TypeVar, Union, cast

import numpy as np
import numpy.typing as npt
from e7awghal import (
    AbstractCapCtrl,
    AbstractCapParam,
    AbstractCapUnit,
    AbstractQuel1Au50Hal,
    AwgCtrl,
    AwgParam,
    AwgUnit,
    CapCtrlSimpleMulti,
    CapIqDataReader,
    ClockcounterCtrl,
    E7awgCaptureDataError,
    E7FwType,
    SimplemultiAwgTriggers,
    SimplemultiSequencer,
    create_quel1au50hal,
)
from e7awghal.common_defs import _DEFAULT_POLLING_PERIOD

logger = logging.getLogger(__name__)


class CaptureReturnCode(Enum):
    CAPTURE_TIMEOUT = 1
    CAPTURE_ERROR = 2
    BROKEN_DATA = 3
    SUCCESS = 4


RT = TypeVar("RT")  # notes: return type
CVRT = TypeVar("CVRT")  # notes: converted return type
CCT = TypeVar("CCT", bound=AbstractCapCtrl)  # notes: capture controller type


class AbstractCancellableTask(
    Generic[RT],
    metaclass=ABCMeta,
):
    __slots__ = {
        "_pool",
        "_args",
        "_request_cancel",
        "_cancel_failed",
        "_cancelled",
        "_future",
        "_lock",
    }

    def __init__(self, pool: ThreadPoolExecutor, *args):
        self._pool = pool
        self._args = args
        self._request_cancel: bool = False
        self._cancel_failed: bool = False
        self._cancelled: bool = False
        self._future: Union[Future[RT], None] = None
        self._lock = RLock()

    @abstractmethod
    def _body(self, *args) -> RT: ...

    def _future_cancel_hook(self) -> None:
        pass

    def _cancellable_wait_loop(
        self,
        check_callback: Callable[[], bool],
        timeout_callback: Callable[[], None],
        cancel_callback: Callable[[], None],
        timeout: Optional[float],
        polling_period: float,
    ) -> None:
        t0 = time.perf_counter()
        while timeout is None or (time.perf_counter() < t0 + timeout):
            time.sleep(polling_period)
            if check_callback():
                break
            if self._request_cancel:
                cancel_callback()
        else:
            timeout_callback()

    def start(self) -> None:
        with self._lock:
            if self._future is None:
                self._future = self._pool.submit(self._body, *self._args)
            else:
                raise RuntimeError("already started")

    def cancel(self, timeout: Optional[float] = None, polling_period: Optional[float] = None) -> bool:
        if polling_period is None:
            if hasattr(self, "_polling_period"):
                polling_period_: float = getattr(self, "_polling_period") or _DEFAULT_POLLING_PERIOD
            else:
                polling_period_ = _DEFAULT_POLLING_PERIOD
        else:
            polling_period_ = polling_period

        if self._future is None:
            raise RuntimeError("not started yet")

        with self._lock:
            self._cancelled = self._future.cancel()
            if not self._cancelled:
                self._request_cancel = True
                t0 = time.perf_counter()
                while timeout is None or time.perf_counter() < t0 + timeout:
                    time.sleep(polling_period_)
                    if isinstance(self._future.exception(), CancelledError):
                        self._cancelled = True
                    if not self._future.running():
                        break
                else:
                    logger.warning(f"failed to stop a task {self.__class__.__name__}({self._args})")
                    self._cancel_failed = True
            else:
                self._future_cancel_hook()

            return self._cancelled

    def result(self, timeout: Optional[float] = None) -> RT:
        if self._future is None:
            raise RuntimeError("not started yet")

        with self._lock:
            if self._cancel_failed:
                if not self.cancel(timeout):
                    raise RuntimeError("cancel request is still pending")
            return self._future.result(timeout)

    def exception(self, timeout: Optional[float] = None) -> Optional[BaseException]:
        if self._future is None:
            raise RuntimeError("not started yet")
        return self._future.exception(timeout)

    def is_started(self) -> bool:
        return self._future is not None

    def running(self) -> bool:
        if self._future:
            return self._future.running()
        else:
            return False

    def done(self) -> bool:
        if self._future:
            # TODO: clarify done() should True or not when canncelled.
            return self._future.done() and not self._cancelled
        else:
            return False

    def cancelled(self) -> bool:
        if self._future:
            return self._future.cancelled() or self._cancelled
        else:
            return False


class AbstractCancellableTaskWrapper(Generic[RT, CVRT], metaclass=ABCMeta):
    def __init__(self, task: AbstractCancellableTask[RT]):
        self._task = task

    def start(self) -> None:
        self._task.start()

    def cancel(self, timeout: Optional[float] = None, polling_period: Optional[float] = None) -> bool:
        return self._task.cancel(timeout, polling_period)

    @abstractmethod
    def _conveter(self, orig: RT) -> CVRT: ...

    def result(self, timeout: Optional[float] = None) -> CVRT:
        return self._conveter(self._task.result(timeout))

    def exception(self, timeout: Optional[float] = None) -> Optional[BaseException]:
        return self._task.exception(timeout)

    def is_started(self) -> bool:
        return self._task.is_started()

    def running(self) -> bool:
        return self._task.running()

    def done(self) -> bool:
        return self._task.done()

    def cancelled(self) -> bool:
        return self._task.cancelled()


class AbstractStartAwgunitsTask(AbstractCancellableTask[None]):
    _DONE_TIMEOUT_MARGIN: Final[float] = 0.25
    _CANCEL_DONE_TIMEOUT: Final[float] = 1.0
    _WORD_RATE: Final[int] = 125_000_000  # [Hz]

    def __init__(
        self,
        pool: ThreadPoolExecutor,
        awgctrl: AwgCtrl,
        awgunits: dict[int, AwgUnit],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ):
        if len(awgunits) == 0:
            raise ValueError("no awg units are specified")
        super().__init__(pool)
        self._awgctrl: AwgCtrl = awgctrl
        self._awgunits: dict[int, AwgUnit] = awgunits
        self._timeout: Union[float, None] = None if disable_timeout else (timeout or self._default_done_timeout)
        self._polling_period: float = polling_period or _DEFAULT_POLLING_PERIOD
        self._has_started_emission: bool = False

    def has_started_emission(self) -> bool:
        return self._has_started_emission

    def wait_for_starting_emission(self) -> None:
        # Notes: usually you don't have to use this method. possible use cases is taking spectrum snapshot with
        #        a spectrum analyzer after confirming RF signal is actually generating.
        #        so, blocking API is enough.
        # Notes: this API should not raise exception since any failures in the main task should be noticed when
        #        either of task.result() or task.exception() is invoked.
        # Notes: this loop should exit in 0.5 (= _PRELOAD_TIMEOUT + _START_TIMEOUT) second or so for TaskNow.
        while not (self._has_started_emission or self.cancelled()):
            if self.done() and not self._has_started_emission:
                # Notes: it should not happen, basically.
                logger.error("emission has not been started due to failures in the task thread")
                break
            time.sleep(self._polling_period)

    @cached_property
    def _hwidxs(self) -> set[int]:
        return {u.unit_index for u in self._awgunits.values()}

    @cached_property
    def _idxs_string(self) -> str:
        return "awgunits" + ", ".join([f"-#{u}" for u in self._awgunits])

    @cached_property
    def _default_done_timeout(self) -> float:
        return (
            max([u.get_wave_duration() for u in self._awgunits.values()]) / self._WORD_RATE + self._DONE_TIMEOUT_MARGIN
        )

    def _clear_done_all(self, raise_exception: bool = False):
        for u in self._awgunits.values():
            u.clear_done()
        if self._awgctrl.are_done_any(self._hwidxs):
            if raise_exception:
                raise RuntimeError(f"failed to clear done flag, some of {self._idxs_string} are still asserted")
            else:
                logger.warning(f"failed to clear done flag, some of {self._idxs_string} are still asserted")

    def _terminate_all(self):
        for u in self._awgunits.values():
            u.terminate(timeout=self._CANCEL_DONE_TIMEOUT, polling_period=self._polling_period).result()

    @abstractmethod
    def _activate_awgunits(self) -> None: ...

    @abstractmethod
    def _wait_for_preparation(self) -> None: ...

    @abstractmethod
    def _wait_for_start(self) -> None: ...

    def _check_start_callback(self) -> bool:
        with self._awgctrl.lock:
            if self._awgctrl.have_started_all(self._hwidxs):
                self._awgctrl.check_error(self._hwidxs)
                return True
            else:
                return False

    def _check_done_callback(self) -> bool:
        with self._awgctrl.lock:
            if self._awgctrl.are_done_all(self._hwidxs):
                self._clear_done_all()
                self._awgctrl.check_error(self._hwidxs)
                return True
            else:
                return False

    def _timeout_callback(self, timeout_msg: str):
        # Notes: don't use hal.awgctrl.terminate(), it wouldn't work well...
        self._terminate_all()
        self._awgctrl.check_error(self._hwidxs)
        raise TimeoutError(timeout_msg)

    def _cancel_callback(self, cancel_msg: str):
        # Notes: don't use hal.awgctrl.terminate(), it wouldn't work well...
        self._terminate_all()
        self._awgctrl.check_error(self._hwidxs)
        raise CancelledError(cancel_msg)

    def _body(self) -> None:
        with self._awgctrl.lock:
            if self._awgctrl.are_busy_any(self._hwidxs):
                raise RuntimeError(f"stop preloading because some of {self._idxs_string} are still busy")
            self._clear_done_all(raise_exception=True)
            self._activate_awgunits()

        self._wait_for_preparation()

        self._wait_for_start()

        self._cancellable_wait_loop(
            self._check_done_callback,
            lambda: self._timeout_callback(f"failed to finish operation of some of {self._idxs_string}"),
            lambda: self._cancel_callback("cancelled during waiting for the completion of the operation"),
            self._timeout,
            self._polling_period,
        )
        return None


class StartAwgunitsNowTask(AbstractStartAwgunitsTask):
    _PRELOAD_TIMEOUT: Final[float] = 0.25
    _START_TIMEOUT: Final[float] = 0.25

    def _activate_awgunits(self) -> None:
        self._awgctrl.prepare(self._hwidxs)

    def _wait_for_preparation(self) -> None:
        self._cancellable_wait_loop(
            self._check_prepare_callback,
            lambda: self._timeout_callback(f"failed to finish preloading at some of {self._idxs_string}"),
            lambda: self._cancel_callback("cancelled during waiting for the completion of preloading"),
            self._PRELOAD_TIMEOUT,
            self._polling_period,
        )

    def _check_prepare_callback(self) -> bool:
        with self._awgctrl.lock:
            if self._awgctrl.are_ready_all(self._hwidxs):
                self._awgctrl.check_error(self._hwidxs)
                return True
            else:
                return False

    def _wait_for_start(self) -> None:
        with self._awgctrl.lock:
            if not self._awgctrl.are_ready_all(self._hwidxs):
                raise RuntimeError(f"stop starting because some of {self._idxs_string} are not ready yet")
            self._awgctrl.start_now(self._hwidxs)

        self._cancellable_wait_loop(
            self._check_start_callback,
            lambda: self._timeout_callback(f"failed to start some of {self._idxs_string}"),
            lambda: self._cancel_callback("cancelled during waiting for starting"),
            self._START_TIMEOUT,
            self._polling_period,
        )

    def _check_start_callback(self) -> bool:
        with self._awgctrl.lock:
            if self._awgctrl.have_started_all(self._hwidxs):
                self._awgctrl.check_error(self._hwidxs)
                self._has_started_emission = True
                return True
            else:
                return False


class StartAwgunitsTimedTask(AbstractStartAwgunitsTask):
    _TRIGGER_SETTABLE_MARGIN: Final[float] = 0.05  # [s]
    _START_TIMEOUT_MARGIN: Final[float] = 0.25  # [s]

    def __init__(
        self,
        pool: ThreadPoolExecutor,
        awgctrl: AwgCtrl,
        awgunits: dict[int, AwgUnit],
        sqrctrl: SimplemultiSequencer,
        clkctrl: ClockcounterCtrl,
        timecount: int,
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ):
        super().__init__(
            pool, awgctrl, awgunits, timeout=timeout, polling_period=polling_period, disable_timeout=disable_timeout
        )
        self._sqrctrl: SimplemultiSequencer = sqrctrl
        self._clkctrl: ClockcounterCtrl = clkctrl
        self._timecount: int = timecount
        self._delta: float = 0.0

    def _activate_awgunits(self) -> None:
        cur, _ = self._clkctrl.read_counter()
        if self._timecount < cur + self._TRIGGER_SETTABLE_MARGIN * self._clkctrl.CLOCK_FREQUENCY:
            raise RuntimeError(f"specified timecount (= {self._timecount}) is too late to schedule")
        self._delta = (self._timecount - cur) / self._clkctrl.CLOCK_FREQUENCY

        trig = SimplemultiAwgTriggers()
        trig.append(self._timecount, self._hwidxs)
        self._sqrctrl.add_awg_start(trig)

    def _wait_for_preparation(self) -> None:
        # do nothing
        pass

    def _wait_for_start(self) -> None:
        self._cancellable_wait_loop(
            self._check_start_callback,
            lambda: self._timeout_before_start_callback(f"failed to start some of {self._idxs_string}"),
            lambda: self._cancel_before_start_callback("cancelled during waiting for starting"),
            self._delta + self._START_TIMEOUT_MARGIN,
            self._polling_period,
        )

    def _check_start_callback(self) -> bool:
        with self._awgctrl.lock:
            if self._awgctrl.have_started_all(self._hwidxs):
                self._awgctrl.check_error(self._hwidxs)
                self._has_started_emission = True
                return True
            else:
                return False

    def _timeout_before_start_callback(self, timeout_msg: str):
        self._sqrctrl.cancel_triggers()
        # Notes: don't use hal.awgctrl.terminate(), it wouldn't work well...
        self._terminate_all()  # Notes: to avoid race condition
        self._awgctrl.check_error(self._hwidxs)
        raise TimeoutError(timeout_msg)

    def _cancel_before_start_callback(self, cancel_msg: str):
        self._sqrctrl.cancel_triggers()
        # Notes: don't use hal.awgctrl.terminate(), it wouldn't work well...
        self._terminate_all()
        self._awgctrl.check_error(self._hwidxs)
        raise CancelledError(cancel_msg)


class AbstractStartCapunitsTask(Generic[CCT], AbstractCancellableTask[dict[tuple[int, int], CapIqDataReader]]):
    # Notes: it is no problem to set long _START_TIMEOUT because upper layer API gives more realistic value as argument.
    _DONE_TIMEOUT_MARGIN: Final[float] = 0.25
    _CANCEL_DONE_TIMEOUT: Final[float] = 1.0
    _WORD_RATE: Final[int] = 125_000_000  # [Hz]

    def __init__(
        self,
        pool: ThreadPoolExecutor,
        capctrl: CCT,
        capunits: dict[tuple[int, int], AbstractCapUnit],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ):
        if len(capunits) == 0:
            raise ValueError("no capture units are specified")
        super().__init__(pool)
        self._capctrl: CCT = capctrl
        self._capunits: dict[tuple[int, int], AbstractCapUnit] = capunits
        self._timeout: Optional[float] = None if disable_timeout else (timeout or self._default_done_timeout)
        self._polling_period: float = polling_period or _DEFAULT_POLLING_PERIOD

    @cached_property
    def _hwidxs(self) -> set[int]:
        return {u.unit_index for u in self._capunits.values()}

    @cached_property
    def _idxs_string(self) -> str:
        return "capunits" + ", ".join([f"-#{m}:{u}" for m, u in self._capunits])

    @cached_property
    def _default_done_timeout(self) -> float:
        return (
            max([u.get_capture_duration() for u in self._capunits.values()]) / self._WORD_RATE
            + self._DONE_TIMEOUT_MARGIN
        )

    def _clear_done_all(self, raise_exception: bool = False):
        for u in self._capunits.values():
            u.clear_done()
        if self._capctrl.are_done_any(self._hwidxs):
            if raise_exception:
                raise RuntimeError(f"failed to clear done flag, some of {self._idxs_string} are still asserted")
            else:
                logger.warning(f"failed to clear done flag, some of {self._idxs_string} are still asserted")

    def _terminate_all(self):
        for i, u in self._capunits.items():
            u.terminate(timeout=self._CANCEL_DONE_TIMEOUT, polling_period=self._polling_period).result()
            u.hard_reset()  # Notes: to flush buffer in FPGA

    @abstractmethod
    def _wait_for_starting(self): ...

    def _check_done_callback(self) -> bool:
        with self._capctrl.lock:
            if self._capctrl.are_done_all(self._hwidxs):
                self._clear_done_all()
                self._capctrl.check_error(self._hwidxs)
                return True
            else:
                return False

    def _timeout_done_callback(self) -> None:
        self._terminate_all()
        self._capctrl.check_error(self._hwidxs)
        raise TimeoutError(f"failed to wait for the completion of {self._idxs_string}")

    @abstractmethod
    def _cancel_callback(self) -> None: ...

    @abstractmethod
    def _activate_capunits(self): ...

    def _body(self) -> dict[tuple[int, int], CapIqDataReader]:
        with self._capctrl.lock:
            for idx, u in self._capunits.items():
                if not u.is_loaded():
                    raise RuntimeError(f"capunit-#{idx} is not configured yet")

            if self._capctrl.are_busy_any(self._hwidxs):
                raise RuntimeError(f"stop starting capture because some of {self._idxs_string} are still busy")
            self._clear_done_all(raise_exception=True)
            self._activate_capunits()

        self._wait_for_starting()

        self._cancellable_wait_loop(
            self._check_done_callback,
            self._timeout_done_callback,
            self._cancel_callback,
            self._timeout,
            self._polling_period,
        )

        readers: dict[tuple[int, int], CapIqDataReader] = {}
        broken_data: bool = False
        for idx, u in self._capunits.items():
            assert u.is_loaded()
            r = u.get_reader()
            n = u.get_num_captured_sample()
            m = r.total_size_in_sample
            if n == m:
                readers[idx] = r
            elif n < m:
                broken_data = True
                logger.error(
                    f"size of the data captured by capunit-#{idx} is shorter than expected "
                    f"({n} samples < {m} samples)"
                )
            else:
                broken_data = True
                logger.error(
                    f"size of the data captured by capunit-#{idx} is longer than expected "
                    f"({n} samples > {m} samples)"
                )

        if broken_data:
            raise E7awgCaptureDataError(f"the captured data is broken at some of {self._idxs_string}")

        return readers


class StartCapunitsNowTask(AbstractStartCapunitsTask[AbstractCapCtrl]):
    def __init__(
        self,
        pool: ThreadPoolExecutor,
        capctrl: AbstractCapCtrl,
        capunits: dict[tuple[int, int], AbstractCapUnit],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ):
        super().__init__(
            pool, capctrl, capunits, timeout=timeout, polling_period=polling_period, disable_timeout=disable_timeout
        )

    def _wait_for_starting(self) -> None:
        pass

    def _cancel_callback(self) -> None:
        self._terminate_all()
        self._capctrl.check_error(self._hwidxs)
        raise CancelledError(f"{self._idxs_string} are stopped during the capture")

    def _activate_capunits(self):
        self._capctrl.start_now(self._hwidxs)


class StartCapunitsByTriggerTask(AbstractStartCapunitsTask[CapCtrlSimpleMulti]):
    _DEFAULT_START_TIMEOUT: Final[float] = 10.0  # [s]

    def __init__(
        self,
        pool: ThreadPoolExecutor,
        capctrl: CapCtrlSimpleMulti,
        capunits: dict[tuple[int, int], AbstractCapUnit],
        *,
        timeout_before_trigger: Optional[float] = None,
        timeout_after_trigger: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ):
        super().__init__(
            pool,
            capctrl,
            capunits,
            timeout=timeout_after_trigger,
            polling_period=polling_period,
            disable_timeout=disable_timeout,
        )
        self._capmod_idxs: set[int] = {cu[0] for cu in capunits}
        self._trigger_activated: bool = False
        self._timeout_before_trigger: Union[float, None] = (
            None if disable_timeout is None else (timeout_before_trigger or self._DEFAULT_START_TIMEOUT)
        )

    def _activate_capunits(self):
        self._capctrl.add_triggerable_units(self._hwidxs)
        self._trigger_activated = True

    def _deactivate_trigger(self):
        if self._trigger_activated:
            self._capctrl.remove_triggerable_units(self._hwidxs)
            for capmod_idx in self._capmod_idxs:
                self._capctrl.set_triggering_awgunit_idx(capmod_idx, None)  # unset
            self._trigger_activated = False

    def _check_start_callback(self) -> bool:
        with self._capctrl.lock:
            has_started: bool = self._capctrl.have_started_all(self._hwidxs)
            has_done: bool = self._capctrl.are_done_all(self._hwidxs)
            if has_started or has_done:
                self._deactivate_trigger()
            return has_started or has_done

    def _timeout_start_callback(self) -> None:
        self._deactivate_trigger()
        self._terminate_all()
        self._capctrl.check_error(self._hwidxs)
        raise TimeoutError(f"failed to wait for the completion of {self._idxs_string}")

    def _wait_for_starting(self) -> None:
        capmod_not_ready: list[int] = []
        for capmod_idx in self._capmod_idxs:
            if self._capctrl.get_triggering_awgunit_idx(capmod_idx) is None:
                capmod_not_ready.append(capmod_idx)
        if len(capmod_not_ready) > 0:
            capmod_str = ", ".join([f"-#{cm}" for cm in capmod_not_ready])
            self._deactivate_trigger()
            self._terminate_all()
            self._capctrl.check_error(self._hwidxs)
            raise RuntimeError(f"no triggering awg unit is set to capture modules{capmod_str}")

        self._cancellable_wait_loop(
            self._check_start_callback,
            self._timeout_start_callback,
            self._cancel_callback,
            self._timeout_before_trigger,
            self._polling_period,
        )

    def _cancel_callback(self) -> None:
        self._deactivate_trigger()
        self._terminate_all()
        self._deactivate_trigger()
        self._capctrl.check_error(self._hwidxs)
        raise CancelledError(f"{self._idxs_string} are stopped during the capture")

    def _future_cancel_hook(self) -> None:
        # Notes: triggering_awgunit is set in advance. need to clear it even if the task is cancelled before start.
        for capmod_idx in self._capmod_idxs:
            self._capctrl.set_triggering_awgunit_idx(capmod_idx, None)


class Quel1WaveSubsystem:
    def __init__(
        self, ipaddr_wss: str, ipaddr_sss: Optional[str] = None, auth_callback: Optional[Callable[[], bool]] = None
    ):
        self._ipaddr_wss: str = ipaddr_wss
        self._ipaddr_sss: str = ipaddr_sss or str(ipaddress.IPv4Address(self._ipaddr_wss) + 0x00010000)
        self._auth_callback: Optional[Callable[[], bool]] = auth_callback
        self._hal: AbstractQuel1Au50Hal = create_quel1au50hal(
            ipaddr_wss=self._ipaddr_wss, ipaddr_sss=self._ipaddr_sss, auth_callback=self._auth_callback
        )

    def initialize(self):
        self.hal.initialize()

    @property
    def ipaddr_wss(self) -> str:
        return self._ipaddr_wss

    @cached_property
    def ipaddr_sss(self) -> str:
        return self._ipaddr_sss

    @property
    def hal(self) -> AbstractQuel1Au50Hal:
        return self._hal

    @cached_property
    def fw_type(self) -> E7FwType:
        return self.hal.fw_type()

    @cached_property
    def fw_version(self) -> str:
        return self.hal.fwversion

    @property
    def num_awgunit(self) -> int:
        return self.hal.awgctrl.num_unit

    @property
    def num_capmod(self) -> int:
        return self.hal.capctrl.num_module

    @property
    def num_capunit(self) -> int:
        return self.hal.capctrl.num_unit

    def num_capunit_of_capmod(self, capmod_idx) -> int:
        return self.hal.capctrl.num_unit_of_module(capmod_idx)

    def _get_awgunit(self, awgunit_idx: int) -> AwgUnit:
        if not (0 <= awgunit_idx < self.num_awgunit):
            raise ValueError(f"invalid index of awgunit: {awgunit_idx}")
        else:
            return self.hal.awgunit(awgunit_idx)

    def _capunit_idx_to_hwidx(self, capunit_idx: tuple[int, int]) -> int:
        m, u = capunit_idx
        return self.hal.capctrl.units_of_module(m)[u]

    def _get_capunit(self, capunit_idx: tuple[int, int]) -> AbstractCapUnit:
        m, u = capunit_idx
        if not ((0 <= m < self.hal.capctrl.num_module) and (0 <= u < self.hal.capctrl.num_unit_of_module(m))):
            raise ValueError(f"invalid index of capunit: {capunit_idx}")
        return self.hal.capunit(self._capunit_idx_to_hwidx(capunit_idx))

    def initialize_sequencer(self):
        self.hal.sqrctrl.initialize()

    def initialize_awgunits(self, awgunit_idxs: Collection[int]):
        for i in awgunit_idxs:
            self._get_awgunit(i).initialize()

    def initialize_all_awgunits(self):
        self.initialize_awgunits(self.hal.awgctrl.units)
        self.hal.awgctrl.initialize()

    def get_current_timecounter(self) -> int:
        return self.hal.clkcntr.read_counter()[0]

    def get_latest_sysref_timecounter(self) -> int:
        cntr = self.hal.clkcntr.read_counter()[1]
        if cntr:
            return cntr
        else:
            raise RuntimeError("installed wss firmware doesn't support SYSREF timecounter")

    def get_averaged_sysref_offset(self, num_iteration: int = 100) -> float:
        cntr = self.hal.clkcntr.read_counter()[1]
        if cntr is None:
            raise RuntimeError("installed wss firmware doesn't support SYSREF timecounter")

        cntr %= 2000
        for _ in range(num_iteration - 1):
            cntr += cast(int, self.hal.clkcntr.read_counter()[1]) % 2000
        return cntr / num_iteration

    def timecounter_to_second(self, tctr: int) -> float:
        return tctr / self.hal.clkcntr.CLOCK_FREQUENCY

    def get_names_of_wavedata(self, awgunit_idx) -> set[str]:
        u = self._get_awgunit(awgunit_idx)
        return u.get_names_of_wavedata()

    def register_wavedata(
        self, awgunit_idx: int, name: str, iq: npt.NDArray[np.complex64], allow_update: bool = True, **kwdargs
    ) -> None:
        u = self._get_awgunit(awgunit_idx)
        if allow_update and u.has_wavedata(name):
            u.delete_wavedata(name)
        u.register_wavedata_from_complex64vector(name, iq, **kwdargs)

    def has_wavedata(self, awgunit_idx: int, label: str) -> bool:
        u = self._get_awgunit(awgunit_idx)
        return u.has_wavedata(label)

    def delete_wavedata(self, awgunit_idx: int, label: str) -> None:
        u = self._get_awgunit(awgunit_idx)
        if u.has_wavedata(label):
            u.delete_wavedata(label)

    def config_awgunit(self, awgunit_idx: int, param: AwgParam):
        u = self._get_awgunit(awgunit_idx)
        u.load_parameter(param)

    def start_awgunits_now(
        self,
        awgunit_idxs: Collection[int],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
        return_after_start_emission: bool = True,
    ) -> StartAwgunitsNowTask:
        awgunits = {idx: self.hal.awgunit(idx) for idx in awgunit_idxs}
        task = StartAwgunitsNowTask(
            self.hal.awgctrl._pool,
            self.hal.awgctrl,
            awgunits,
            timeout=timeout,
            polling_period=polling_period,
            disable_timeout=disable_timeout,
        )
        task.start()
        if return_after_start_emission:
            task.wait_for_starting_emission()
        return task

    def start_awgunits_timed(
        self,
        awgunit_idxs: Collection[int],
        timecount: int,
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
        return_after_start_emission: bool = False,
    ) -> StartAwgunitsTimedTask:
        awgunits = {idx: self.hal.awgunit(idx) for idx in awgunit_idxs}
        task = StartAwgunitsTimedTask(
            self.hal.awgctrl._pool,
            self.hal.awgctrl,
            awgunits,
            self.hal.sqrctrl,
            self.hal.clkcntr,
            timecount=timecount,
            timeout=timeout,
            polling_period=polling_period,
            disable_timeout=disable_timeout,
        )
        task.start()
        if return_after_start_emission:
            task.wait_for_starting_emission()
        return task

    def initialize_capunits(self, capunit_idxs: Collection[tuple[int, int]]):
        for i in capunit_idxs:
            self._get_capunit(i).initialize()

    def initialize_all_capunits(self):
        cc = self.hal.capctrl
        if isinstance(cc, CapCtrlSimpleMulti):
            for i in cc.modules:
                cc.set_triggering_awgunit_idx(i, None)
        for i in self.hal.capctrl.units:
            self.hal.capunit(i).initialize()

    def config_capunit(self, capunit_idx: tuple[int, int], param: AbstractCapParam):
        u = self._get_capunit(capunit_idx)
        u.load_parameter(param)

    def start_capunits_now(
        self,
        capunits_idxs: Collection[tuple[int, int]],
        *,
        timeout: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ) -> StartCapunitsNowTask:
        capunits = {idx: self.hal.capunit(self._capunit_idx_to_hwidx(idx)) for idx in capunits_idxs}
        task = StartCapunitsNowTask(
            self.hal.capctrl._pool,
            self.hal.capctrl,
            capunits,
            timeout=timeout,
            polling_period=polling_period,
            disable_timeout=disable_timeout,
        )
        task.start()
        return task

    def set_triggering_awg_to_line(self, capmod_idx: int, awg_idx: int):
        cc = self.hal.capctrl
        if isinstance(cc, CapCtrlSimpleMulti):
            cc.set_triggering_awgunit_idx(capmod_idx, awg_idx)
        else:
            raise TypeError(f"the specified capture controller {cc.__class__.__name__} is not supported")

    def unset_triggering_awg_to_line(self, capmod_idx: int):
        cc = self.hal.capctrl
        if isinstance(cc, CapCtrlSimpleMulti):
            cc.set_triggering_awgunit_idx(capmod_idx, None)
        else:
            raise TypeError(f"the specified capture controller {cc.__class__.__name__} is not supported")

    def get_triggering_awg_to_line(self, capmod_idx: int) -> Union[int, None]:
        cc = self.hal.capctrl
        if isinstance(cc, CapCtrlSimpleMulti):
            return cc.get_triggering_awgunit_idx(capmod_idx)
        else:
            raise TypeError(f"the specified capture controller {cc.__class__.__name__} is not supported")

    def start_capunits_by_trigger(
        self,
        capunits_idxs: Collection[tuple[int, int]],
        *,
        timeout_before_trigger: Optional[float] = None,
        timeout_after_trigger: Optional[float] = None,
        polling_period: Optional[float] = None,
        disable_timeout: bool = False,
    ) -> StartCapunitsByTriggerTask:
        capunits = {idx: self.hal.capunit(self._capunit_idx_to_hwidx(idx)) for idx in capunits_idxs}
        cc = self.hal.capctrl
        if isinstance(cc, CapCtrlSimpleMulti):
            task = StartCapunitsByTriggerTask(
                self.hal.capctrl._pool,
                cc,
                capunits,
                timeout_before_trigger=timeout_before_trigger,
                timeout_after_trigger=timeout_after_trigger,
                polling_period=polling_period,
                disable_timeout=disable_timeout,
            )
            task.start()
            return task
        else:
            raise TypeError(f"the specified capture controller {cc.__class__.__name__} is not supported")



import logging
from typing import Dict

from quel_ic_config.abstract_ic import AbstractIcConfigHelper, AbstractIcMixin, Gpio6

logger = logging.getLogger(__name__)


PathselectorboardGpioRegs: Dict[int, type] = {
    0: Gpio6,
}


PathselectorboardGpioRegNames: Dict[str, int] = {
    "RFSW": 0,
}


class PathselectorboardGpioMixin(AbstractIcMixin):
    Regs = PathselectorboardGpioRegs
    RegNames = PathselectorboardGpioRegNames

    def __init__(self, name):
        super().__init__(name)

    def dump_regs(self) -> Dict[int, int]:
        """dump all the available registers.
        :return: a mapping between an address and a value of the registers
        """
        regs = {}
        for addr in self.Regs:
            regs[addr] = self.read_reg(addr)  # actually addr is ignored by exstickge
        return regs


class PathselectorboardGpioConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming GPIO with convenient notations. It also provides caching capability that
    keep modifications on the registers to write them at once with flash_updated() in a right order.
    """

    def __init__(self, ic: PathselectorboardGpioMixin):
        super().__init__(ic)

    def flush(self, discard_after_flush=True):
        for addr in self.ic.Regs:
            if addr in self.updated:
                self.ic.write_reg(addr, self.updated[addr])
        if discard_after_flush:
            self.discard()



import logging
from dataclasses import dataclass, field
from typing import Dict, cast

from quel_ic_config.abstract_ic import AbstractIcConfigHelper, AbstractIcMixin, AbstractIcReg, b_1bf_bool, p_1bf_bool

logger = logging.getLogger(__name__)


@dataclass
class RfSwitchArrayReg(AbstractIcReg):
    path0: bool = field(default=False)
    path1: bool = field(default=False)
    path2: bool = field(default=False)
    path3: bool = field(default=False)
    monitor: bool = field(default=False)

    def _parse_switch_pair(self, r):
        if r == 0x00:
            return False
        elif r == 0x03:
            return True
        elif r == 0x01 or r == 0x02:
            logger.warning("invalid state of RF switch for loopback, considered as inside")
            return True
        else:
            raise AssertionError


class AbstractRfSwitchArrayMixin(AbstractIcMixin):
    def dump_regs(self) -> Dict[int, int]:
        """dumping all the available registers.

        :return: a mapping between an address and a value of the registers
        """
        regs = {0: self.read_reg(0)}
        return regs


class QubeRfSwitchArray0Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = self._parse_switch_pair(v & 0b00000000000011)
        self.path1 = p_1bf_bool(v, 2)
        self.monitor = self._parse_switch_pair((v & 0b00000000011000) >> 3)
        self.path2 = p_1bf_bool(v, 5)
        self.path3 = p_1bf_bool(v, 6)

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 0)
            | b_1bf_bool(self.path0, 1)
            | b_1bf_bool(self.path1, 2)
            | b_1bf_bool(self.monitor, 3)
            | b_1bf_bool(self.monitor, 4)
            | b_1bf_bool(self.path2, 5)
            | b_1bf_bool(self.path3, 6)
        )


class QubeRfSwitchArray1Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = self._parse_switch_pair((v & 0b11000000000000) >> 12)
        self.path1 = p_1bf_bool(v, 11)
        self.monitor = self._parse_switch_pair((v & 0b00011000000000) >> 9)
        self.path2 = p_1bf_bool(v, 8)
        self.path3 = p_1bf_bool(v, 7)

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 13)
            | b_1bf_bool(self.path0, 12)
            | b_1bf_bool(self.path1, 11)
            | b_1bf_bool(self.monitor, 10)
            | b_1bf_bool(self.monitor, 9)
            | b_1bf_bool(self.path2, 8)
            | b_1bf_bool(self.path3, 7)
        )


QubeRfSwitchRegs: Dict[int, type] = {
    0: QubeRfSwitchArray0Reg,
    1: QubeRfSwitchArray1Reg,
}

QubeSwitchRegNames: Dict[str, int] = {
    "Group0": 0,
    "Group1": 1,
}


class QubeRfSwitchArrayMixin(AbstractRfSwitchArrayMixin):
    Regs = QubeRfSwitchRegs
    RegNames = QubeSwitchRegNames

    def __init__(self, name):
        super().__init__(name)


class Quel1TypeARfSwitchArray0Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = p_1bf_bool(v, 0)  # port0: read-in,  port1: read-out
        self.path1 = p_1bf_bool(v, 5)  # port3 for typeA, port2 for TypeB
        self.monitor = p_1bf_bool(v, 3)  # port5: monitor-in, port6: monitor-out
        self.path2 = p_1bf_bool(v, 2)  # port2 for typeA, port3 for TypeB
        self.path3 = p_1bf_bool(v, 6)  # port4

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 0)
            | b_1bf_bool(self.path1, 5)
            | b_1bf_bool(self.monitor, 3)
            | b_1bf_bool(self.path2, 2)
            | b_1bf_bool(self.path3, 6)
        )


class Quel1TypeARfSwitchArray1Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = p_1bf_bool(v, 12)  # port7: readin, port8: read-out
        self.path1 = p_1bf_bool(v, 8)  # port10 for typeA, port9 for typeB
        self.monitor = p_1bf_bool(v, 9)  # port12: monitor-in, port13: monitor-out
        self.path2 = p_1bf_bool(v, 7)  # port9 for typeA, port10 for typeB
        self.path3 = p_1bf_bool(v, 11)  # port11

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 12)
            | b_1bf_bool(self.path1, 8)
            | b_1bf_bool(self.monitor, 9)
            | b_1bf_bool(self.path2, 7)
            | b_1bf_bool(self.path3, 11)
        )


class Quel1TypeBRfSwitchArray0Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = p_1bf_bool(v, 0)  # port0: read-in,  port1: read-out
        self.path1 = p_1bf_bool(v, 2)  # port3 for typeA, port2 for TypeB
        self.monitor = p_1bf_bool(v, 3)  # port5: monitor-in, port6: monitor-out
        self.path2 = p_1bf_bool(v, 5)  # port2 for typeA, port3 for TypeB
        self.path3 = p_1bf_bool(v, 6)  # port4

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 0)
            | b_1bf_bool(self.path1, 2)
            | b_1bf_bool(self.monitor, 3)
            | b_1bf_bool(self.path2, 5)
            | b_1bf_bool(self.path3, 6)
        )


class Quel1TypeBRfSwitchArray1Reg(RfSwitchArrayReg):
    def parse(self, v: int) -> None:
        self.path0 = p_1bf_bool(v, 12)  # port7: readin, port8: read-out
        self.path1 = p_1bf_bool(v, 11)  # port10 for typeA, port9 for typeB
        self.monitor = p_1bf_bool(v, 9)  # port12: monitor-in, port13: monitor-out
        self.path2 = p_1bf_bool(v, 7)  # port9 for typeA, port10 for typeB
        self.path3 = p_1bf_bool(v, 8)  # port11

    def build(self) -> int:
        return (
            b_1bf_bool(self.path0, 12)
            | b_1bf_bool(self.path1, 11)
            | b_1bf_bool(self.monitor, 9)
            | b_1bf_bool(self.path2, 7)
            | b_1bf_bool(self.path3, 8)
        )


Quel1TypeARfSwitchRegs: Dict[int, type] = {
    0: Quel1TypeARfSwitchArray0Reg,
    1: Quel1TypeARfSwitchArray1Reg,
}

Quel1TypeBRfSwitchRegs: Dict[int, type] = {
    0: Quel1TypeBRfSwitchArray0Reg,
    1: Quel1TypeBRfSwitchArray1Reg,
}

Quel1SwitchRegNames: Dict[str, int] = {
    "Group0": 0,
    "Group1": 1,
}


class Quel1TypeARfSwitchArrayMixin(AbstractRfSwitchArrayMixin):
    Regs = Quel1TypeARfSwitchRegs
    RegNames = Quel1SwitchRegNames

    def __init__(self, name):
        super().__init__(name)


class Quel1TypeBRfSwitchArrayMixin(AbstractRfSwitchArrayMixin):
    Regs = Quel1TypeBRfSwitchRegs
    RegNames = Quel1SwitchRegNames

    def __init__(self, name):
        super().__init__(name)


class RfSwitchArrayConfigHelper(AbstractIcConfigHelper):
    """Helper class for programming RF Switch Array of Qube and Quel-1."""

    # TODO: consider to introduce general mechanisms for supporting virtual registers.
    #       (paged registers may be managed with the mechanisms well.)
    def __init__(self, ic: AbstractRfSwitchArrayMixin):
        super().__init__(ic)

    def _read_reg(self, addr: int, refer_to_cache: bool = False) -> AbstractIcReg:
        data = self.ic.Regs[addr]()
        if addr in self.updated and refer_to_cache:
            data.parse(self.updated[addr])
        else:
            v = self.ic.read_reg(0)
            if addr == 0:
                data.parse(v & 0b00000001111111)
            elif addr == 1:
                data.parse(v & 0b11111110000000)
            else:
                raise ValueError(f"Invalid address: {addr}")

        return cast(AbstractIcReg, data)

    def flush(self, discard_after_flush=True):
        # notes: load all the virtual registers
        if 0 not in self.updated or 1 not in self.updated:
            if 0 not in self.updated:
                v0 = self._read_reg(0, True).build()
                self.updated[0] = v0 & 0b00000001111111
            if 1 not in self.updated:
                v1 = self._read_reg(1, True).build()
                self.updated[1] = v1 & 0b11111110000000

        self.ic.write_reg(0, self.updated[0] | self.updated[1])  # address is not used, actually.
        if discard_after_flush:
            self.discard()
